import logging
from typing import Any, Literal
from fastapi import FastAPI, HTTPException
from fastapi.responses import JSONResponse
from fastapi.middleware.cors import CORSMiddleware
from pydantic import BaseModel
from src.lpsim import Match, Deck
from tests.utils_for_test import get_random_state, set_16_omni, make_respond
from src.lpsim.agents import NothingAgent, InteractionAgent


HTTPServer = FastAPI()
logging.basicConfig(level = logging.INFO)


default_deck_str = '''
        default_version:4.1
        charactor:Kaedehara Kazuha
        charactor:Klee
        charactor:Kaeya
        Dunyarzad*10
        Chef Mao*10
        Timmie*10
        Sweet Madame*10
'''
deck_str_1 = '''
charactor:Fischl@3.3
charactor:Rhodeia of Loch@3.3
charactor:Fatui Pyro Agent@3.3
Tubby@3.3
Sumeru City@3.7
Lucky Dog's Silver Circlet@3.3
Strategize@3.3
Leave It to Me!@3.3
Paid in Full@3.3
Sangonomiya Shrine@3.7
Lotus Flower Crisp@3.3
Mushroom Pizza@3.3
Favonius Cathedral@3.3
Mushroom Pizza@3.3
Magic Guide@3.3*19
'''
deck_str_2 = '''
charactor:Nahida@3.7
charactor:Rhodeia of Loch@3.3
charactor:Fischl@3.3
Gambler's Earrings@3.8
Paimon@3.3
Chef Mao@4.1
Dunyarzad@4.1
The Bestest Travel Companion!@3.3
Paimon@3.3
Liben@3.3
Liben@3.3
Send Off@3.7
Teyvat Fried Egg@4.1
Dunyarzad@4.1
Sweet Madame@3.3
I Haven't Lost Yet!@4.0
Send Off@3.7
Lotus Flower Crisp@3.3
Magic Guide@3.3*15
'''
# deck_str_1 = default_deck_str
# deck_str_2 = default_deck_str
match: Match = Match()
decks: list[Deck] = [Deck.from_str(deck_str_1), 
                     Deck.from_str(deck_str_2)]
# change HP
# for charactor in deck.charactors:
#     charactor.max_hp = 30
#     charactor.hp = 30


# Add the CORSMiddleware to the app
HTTPServer.add_middleware(
    CORSMiddleware,
    allow_origins=[
        'http://localhost:4000', 'https://localhost:4000', 
        'http://127.0.0.1:4000', 'https://127.0.0.1:4000',
    ],
    # allow_origins=['*'],
    allow_credentials=True,
    allow_methods=['*'],
    allow_headers=['*'],
)


def get_new_match(seed: Any = None, rich: bool = False):
    if seed:
        match: Match = Match(random_state = seed)
    else:
        match: Match = Match()
    match.set_deck(decks)
    match.config.max_same_card_number = None
    match.config.charactor_number = None
    match.config.card_number = None
    match.config.random_first_player = False
    match.config.history_level = 10
    match.config.make_skill_prediction = True
    # match.config.initial_hand_size = 20
    # match.config.max_hand_size = 30
    match.config.check_deck_restriction = False
    match.config.recreate_mode = True
    match.config.random_object_information = {
        'rhodeia': ['squirrel', 'raptor', 'frog', 'squirrel', 'raptor', 'frog',
                    'frog', 'squirrel', 'squirrel']
    }
    match.config.player_go_first = 1
    if rich:
        set_16_omni(match)
    match.start()
    match._save_history()
    match.step()
    if agent_0.__class__ != InteractionAgent:
        while match.need_respond(0):
            make_respond(agent_0, match)
    return match


agent_0 = NothingAgent(player_idx = 0)
agent_0 = InteractionAgent(player_idx = 0, only_use_command = True)
agent_1 = InteractionAgent(player_idx = 1, only_use_command = True)


@HTTPServer.on_event('startup')
async def startup_event():
    global match
    match = get_new_match(seed = get_random_state(), rich = False)


class ResetData(BaseModel):
    fixed_random_seed: bool = False
    offset: int = 0
    rich_mode: bool = False
    match_state: Match | None = None
    match_state_idx: int | None = None


@HTTPServer.post('/reset')
async def reset(data: ResetData):
    """
    Reset the match. 
    If match_state_idx is not None, the match will be reset to the
    state of match_state_idx in history. This will keep histories before
    match_state_idx.
    Otherwise, if match_state is not None, the match will be reset to
    match_state. This will clear all histories and start with zero.
    Otherwise, a new match will be created, with 
    fixed_random_seed, offset and rich_mode.

    Return: The dict contains last state with its idx of the match if success.
    """
    global match
    fixed_random_seed = data.fixed_random_seed
    offset = data.offset
    rich = data.rich_mode
    match_state = data.match_state
    match_state_idx = data.match_state_idx
    if match_state_idx is not None:
        history = match._history[:]
        history_diff = match._history_diff[:]
        if len(history) <= match_state_idx or match_state_idx < 0:
            raise HTTPException(status_code = 404, detail = 'State not found')
        match = history[match_state_idx].copy(deep = True)
        match._history = history[:match_state_idx + 1]
        match._history_diff = history_diff[:match_state_idx + 1]
    elif match_state is not None:
        match = match_state
        match._save_history()
    else:
        if fixed_random_seed:
            random_seed = get_random_state(offset)
        else:
            random_seed = None
        match = get_new_match(seed = random_seed, rich = rich)
    return {
        'idx': match_state_idx,
        'match': match.dict(),
        'type': 'FULL',
    }


@HTTPServer.get('/deck/{player_idx}')
async def get_deck(player_idx: int):
    """
    get deck of a player.
    """
    if player_idx < 0 or player_idx > 1:
        raise HTTPException(status_code = 404, detail = 'Player not found')
    return decks[player_idx].dict()


@HTTPServer.get('/decks')
async def get_decks():
    """
    get decks of both players.
    """
    return [deck.dict() for deck in decks]


class DeckData(BaseModel):
    deck_str: str
    player_idx: int


@HTTPServer.post('/deck')
async def post_deck(data: DeckData):
    """
    Set deck.
    """
    deck_str = data.deck_str
    player_idx = data.player_idx
    if player_idx < 0 or player_idx > 1:
        raise HTTPException(status_code = 404, detail = 'Player not found')
    deck = Deck.from_str(deck_str)
    # if not match.state not in [MatchState.WAITING, 
    #                            MatchState.ENDED, 
    #                            MatchState.ERROR]:
    #     raise HTTPException(status_code = 403, detail = 'Match is running')
    if not deck.check_legal(
        match.config.card_number, match.config.max_same_card_number, 
        match.config.charactor_number, match.config.check_deck_restriction
    ):
        raise HTTPException(status_code = 403, detail = 'Deck not legal')
    global decks
    decks[player_idx] = deck


@HTTPServer.get('/state/{mode}/{state_idx}/{player_idx}')
async def get_game_state(
    mode: Literal['one', 'after'], state_idx: int, player_idx: int
):
    """
    Return list of state and its index.

    mode is one: only get the state of state_idx (but also wrapped in a list)
    mode is after: get the state of state_idx and all states after it

    state_idx: -1 means the last state; otherwise, it is the index of the state

    player_idx is -1: fetch complete data
    player_idx is 0 or 1: fetch data for player idx (currently not implemented)
    """
    # player idx check
    if player_idx < -1 or player_idx > 1:
        raise HTTPException(status_code = 404, detail = 'Player not found')
    if player_idx != -1:
        raise HTTPException(status_code = 404, 
                            detail = 'player data fetch not suppoted')
    # state idx check
    if state_idx < -1 or state_idx > len(match._history):
        raise HTTPException(status_code = 404, detail = 'State not found')
    if state_idx == -1:
        state_idx = len(match._history) - 1
    if state_idx == len(match._history):
        # ask for the state after the last state
        return JSONResponse([])
    result = [{
        'idx': state_idx,
        'match': match._history[state_idx].dict(),
        'type': 'FULL',
    }]
    if mode == 'after':
        result += [{
            'idx': state_idx + 1 + idx,
            'match_diff': diff,
            'type': 'DIFF',
        } for idx, diff in enumerate(match._history_diff[state_idx + 1:])]
    print('state output length', len(result))
    return JSONResponse(result)


@HTTPServer.get('/request/{player_idx}')
async def get_request(player_idx: int):
    """
    Return the requests of player_idx. if player_idx is -1, return all 
    requests.
    """
    if player_idx < -1 or player_idx > 1:
        raise HTTPException(status_code = 404, detail = 'Player not found')
    res = match.requests
    if player_idx != -1:
        res = [x for x in res if x.player_idx == player_idx]
    return JSONResponse([x.dict() for x in res])


class RespondData(BaseModel):
    player_idx: int
    command: str


@HTTPServer.post('/respond')
async def post_respond(data: RespondData):
    player_idx = data.player_idx
    command = data.command
    current_history_length = len(match._history)
    if not match.need_respond(player_idx):
        raise HTTPException(status_code = 404, detail = 'Not your turn')
    if player_idx == 0:
        agent = agent_0
    elif player_idx == 1:
        agent = agent_1
    else:
        raise HTTPException(status_code = 404, detail = 'Player not found')
    if agent.__class__ == NothingAgent:
        raise HTTPException(status_code = 404, 
                            detail = 'Cannot control this agent')
    agent.commands = [command]
    resp = agent.generate_response(match)
    if resp is None:
        raise HTTPException(status_code = 404, detail = 'Invalid command')
    match.respond(resp)
    match.step()
    for agent in (agent_0, agent_1):
        if (
            agent.__class__ != InteractionAgent 
            or len(agent.commands) > 0  # type: ignore
        ):
            while match.need_respond(agent.player_idx):
                make_respond(agent, match)
    ret = []
    for idx, [state, diff] in enumerate(zip(
            match._history[current_history_length:],
            match._history_diff[current_history_length:],)):
        if idx == 0:
            ret.append({
                'idx': idx + current_history_length,
                'match': state.dict(),
                'type': 'FULL',
            })
        else:
            ret.append({
                'idx': idx + current_history_length,
                'match_diff': diff,
                'type': 'DIFF',
            })
    return JSONResponse(ret)



from typing import Any, List, Literal
from ...action import Actions
from ..charactor_base import ElementalNormalAttackBase
from ..hydro.sangonomiya_kokomi import SangonomiyaKokomi as SK_3_6
from ..hydro.sangonomiya_kokomi import NereidsAscension as NA_3_6
from ..hydro.sangonomiya_kokomi import KuragesOath


class NereidsAscension(NA_3_6):
    desc: str = (
        'Deals 3 Hydro DMG. This character gains Ceremonial Garment.'
    )
    damage: int = 3

    def get_actions(self, match: Any) -> List[Actions]:
        """
        No healing
        """
        return super(NA_3_6, self).get_actions(match) + [
            self.create_charactor_status('Ceremonial Garment'),
        ]


class SangonomiyaKokomi_3_5(SK_3_6):
    version: Literal['3.5']
    skills: List[
        ElementalNormalAttackBase | KuragesOath | NereidsAscension
    ] = []

    def _init_skills(self) -> None:
        self.skills = [
            ElementalNormalAttackBase(
                name = 'The Shape of Water',
                damage_type = self.element,
                cost = ElementalNormalAttackBase.get_cost(self.element),
            ),
            KuragesOath(),
            NereidsAscension(),
        ]



from typing import Any, List, Literal

from ...event import CreateObjectEventArguments, GameStartEventArguments
from ...action import Actions, ChangeObjectUsageAction, CreateObjectAction
from ...struct import Cost

from ...consts import (
    DamageElementalType, DieColor, ElementType, FactionType, 
    ObjectPositionType, WeaponType
)
from ..charactor_base import (
    ElementalBurstBase, ElementalSkillBase, 
    PhysicalNormalAttackBase, PassiveSkillBase, CharactorBase, SkillTalent
)


# Skills


class Prowl(ElementalSkillBase):
    name: Literal['Prowl'] = 'Prowl'
    desc: str = '''Deals 1 Pyro DMG. This character gains Stealth.'''
    damage: int = 1
    damage_type: DamageElementalType = DamageElementalType.PYRO
    cost: Cost = Cost(
        elemental_dice_color = DieColor.PYRO,
        elemental_dice_number = 3
    )

    def get_actions(self, match: Any) -> List[Actions]:
        """
        Attack and create object
        """
        return super().get_actions(match) + [
            self.create_charactor_status('Stealth')
        ]


class StealthMaster(PassiveSkillBase):
    name: Literal['Stealth Master'] = 'Stealth Master'
    desc: str = (
        '(Passive) When the battle begins, this character gains Stealth.'
    )

    def event_handler_GAME_START(
        self, event: GameStartEventArguments, match: Any
    ) -> List[CreateObjectAction]:
        """
        When game begin, gain stealth
        """
        return [self.create_charactor_status('Stealth')]


# Talents


class PaidinFull(SkillTalent):
    name: Literal['Paid in Full']
    desc: str = (
        'Combat Action: When your active character is Fatui Pyro Agent, '
        'equip this card. After Fatui Pyro Agent equips this card, '
        'immediately use Prowl once. When your Fatui Pyro Agent, who has '
        'this card equipped, creates a Stealth, it will have the following '
        'effect: Starting Usage(s) +1, the Physical DMG the character deals '
        'will be converted to Pyro DMG.'
    )
    version: Literal['3.3'] = '3.3'
    charactor_name: Literal['Fatui Pyro Agent'] = 'Fatui Pyro Agent'
    cost: Cost = Cost(
        elemental_dice_color = DieColor.PYRO,
        elemental_dice_number = 3
    )
    skill: Prowl = Prowl()

    def event_handler_CREATE_OBJECT(
        self, event: CreateObjectEventArguments, match: Any
    ) -> List[ChangeObjectUsageAction]:
        """
        When create object, check if it is stealth
        """
        if self.position.area != ObjectPositionType.CHARACTOR:
            # not equipped, do nothing
            return []
        if not self.position.check_position_valid(
            event.action.object_position, match, player_idx_same = True,
            charactor_idx_same = True, 
            target_area = ObjectPositionType.CHARACTOR_STATUS
        ):
            # not self create charactor status, do nothing
            return []
        if event.action.object_name != 'Stealth':
            # not stealth, do nothing
            return []
        charactor = match.player_tables[
            self.position.player_idx].charactors[self.position.charactor_idx]
        status = charactor.status[event.create_idx]
        assert status.name == 'Stealth'
        # add 1 usage
        return [ChangeObjectUsageAction(
            object_position = status.position,
            change_type = 'DELTA',
            change_usage = 1
        )]

# charactor base


class FatuiPyroAgent(CharactorBase):
    name: Literal['Fatui Pyro Agent']
    version: Literal['3.3'] = '3.3'
    desc: str = '''"Blade of Settlement" Agent'''
    element: ElementType = ElementType.PYRO
    max_hp: int = 10
    max_charge: int = 2
    skills: List[
        PhysicalNormalAttackBase | Prowl | ElementalBurstBase | StealthMaster
    ] = []
    faction: List[FactionType] = [
        FactionType.FATUI
    ]
    weapon_type: WeaponType = WeaponType.OTHER
    talent: PaidinFull | None = None

    def _init_skills(self) -> None:
        self.skills = [
            PhysicalNormalAttackBase(
                name = 'Thrust',
                cost = PhysicalNormalAttackBase.get_cost(self.element),
            ),
            Prowl(),
            ElementalBurstBase(
                name = 'Blade Ablaze',
                damage = 5,
                damage_type = DamageElementalType.PYRO,
                cost = ElementalBurstBase.get_cost(self.element, 3, 2),
            ),
            StealthMaster()
        ]



from .agent_base import AgentBase
from ..server.match import Match
from ..server.interaction import (
    Responses,
    SwitchCardResponse,
    ChooseCharactorResponse,
    RerollDiceResponse,
    DeclareRoundEndResponse
)


class NothingAgent(AgentBase):
    """
    Agent that do nothing, only response essential requests.
    """
    def generate_response(self, match: Match) -> Responses | None:
        for req in match.requests:
            if req.player_idx == self.player_idx:
                if req.name == 'SwitchCardRequest':
                    return SwitchCardResponse(
                        request = req, card_idxs = []
                    )
                elif req.name == 'ChooseCharactorRequest':
                    return ChooseCharactorResponse(
                        request = req, 
                        charactor_idx = req.available_charactor_idxs[0]
                    )
                elif req.name == 'RerollDiceRequest':
                    return RerollDiceResponse(
                        request = req, reroll_dice_idxs = []
                    )
                elif req.name == 'DeclareRoundEndRequest':
                    return DeclareRoundEndResponse(
                        request = req
                    )



import pydantic
from typing import List, get_type_hints


class BaseModel(pydantic.BaseModel):
    class Config:
        extra = pydantic.Extra.forbid  # default forbid extra fields


def list_unique_range_right(data: List[int], minn: int, maxn: int) -> bool:
    """
    Check if the list is unique and in range [minn, maxn).
    """
    if len(data) != len(set(data)):
        return False
    for i in data:
        if i < minn or i >= maxn:
            return False
    return True


"""
parse instance in unions. as hundreds of classes are included in the union,
to accelerate parse speed, we split the union into several unions by their
name.
"""


# key: a union. value: a dict. key: name literal. value: class unions.
_union_dict = {}


def _create_new_dict(types, key: str):
    new_dict = {}
    for type in types.__args__:
        name_hints = get_type_hints(type)[key]
        for name in name_hints.__args__:
            if name not in new_dict:
                new_dict[name] = type
            new_dict[name] |= type
    _union_dict[types] = new_dict


def get_instance_from_type_unions(types, args, key = 'name'):
    """
    Get instance from type unions.
    """
    if types not in _union_dict:
        # print('new', args['name'])
        _create_new_dict(types, key)
    target_types = _union_dict[types][args[key]]
    return pydantic.parse_obj_as(target_types, args)


if __name__ == "__main__":
    pass



from typing import Any, List, Literal

from ...summon.base import SwirlChangeSummonBase

from ...modifiable_values import DamageIncreaseValue

from ...action import ActionTypes, Actions
from ...struct import Cost

from ...consts import (
    DamageElementalType, DieColor, ElementType, FactionType, WeaponType
)
from ..charactor_base import (
    ElementalBurstBase, ElementalNormalAttackBase, ElementalSkillBase, 
    CharactorBase, SkillTalent
)


# Summons


class LargeWindSpirit(SwirlChangeSummonBase):
    name: Literal['Large Wind Spirit'] = 'Large Wind Spirit'
    desc: str = (
        'End Phase: Deal 2 Anemo DMG. '
        'After your character or Summon triggers a Swirl reaction: Convert '
        'the Elemental Type of this card and change its DMG dealt to the '
        'element Swirled. (Can only be converted once before leaving the '
        'field)'
    )
    talent_desc: str = (
        '. When this card has converted to another Elemental Type: Deal +1 '
        'DMG of this Elemental Type.'
    )
    version: Literal['3.3'] = '3.3'
    usage: int = 3
    max_usage: int = 3
    damage: int = 2
    talent_activated: bool = False

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        if self.talent_activated:
            self.desc = self.desc.replace(self.talent_desc, '')
            self.desc += self.talent_desc

    def renew(self, obj: 'LargeWindSpirit') -> None:
        super().renew(obj)
        if obj.talent_activated and not self.talent_activated:
            self.desc += self.talent_desc
            self.talent_activated = obj.talent_activated

    def value_modifier_DAMAGE_INCREASE(
        self, value: DamageIncreaseValue, match: Any,
        mode: Literal['TEST', 'REAL'],
    ) -> DamageIncreaseValue:
        """
        If talent activated, and damage type is not anemo, increase damage
        """
        if (
            not self.talent_activated 
            or self.damage_elemental_type == DamageElementalType.ANEMO
        ):
            # no talent or current is anemo
            return value
        if (
            value.position.player_idx != self.position.player_idx
            or value.target_position.player_idx == self.position.player_idx
        ):
            # enemy or attack self
            return value
        if value.damage_elemental_type != self.damage_elemental_type:
            # elemental type not match
            return value
        # increase damage
        assert mode == 'REAL'
        value.damage += 1
        return value


# Skills


class AstableAnemohypostasisCreation6308(ElementalSkillBase):
    name: Literal[
        'Astable Anemohypostasis Creation - 6308'
    ] = 'Astable Anemohypostasis Creation - 6308'
    desc: str = (
        'Deals 3 Anemo DMG, the target is forcibly switched to the previous '
        'character.'
    )
    damage: int = 3
    damage_type: DamageElementalType = DamageElementalType.ANEMO
    cost: Cost = Cost(
        elemental_dice_color = DieColor.ANEMO,
        elemental_dice_number = 3
    )

    def get_actions(self, match: Any) -> List[Actions]:
        ret = super().get_actions(match)
        prev_idx = match.player_tables[
            1 - self.position.player_idx].previous_charactor_idx()
        if prev_idx is None:
            return ret
        # change charactor
        attack_action = ret[0]
        assert attack_action.type == ActionTypes.MAKE_DAMAGE
        attack_action.do_charactor_change = True
        attack_action.charactor_change_idx = prev_idx
        return ret


class ForbiddenCreationIsomer75TypeII(ElementalBurstBase):
    name: Literal[
        'Forbidden Creation - Isomer 75 / Type II'
    ] = 'Forbidden Creation - Isomer 75 / Type II'
    desc: str = 'Deals 1 Anemo DMG, summons 1 Large Wind Spirit.'
    damage: int = 1
    damage_type: DamageElementalType = DamageElementalType.ANEMO
    cost: Cost = Cost(
        elemental_dice_color = DieColor.ANEMO,
        elemental_dice_number = 3,
        charge = 2
    )

    def get_actions(self, match: Any) -> List[Actions]:
        args = { 'talent_activated': self.is_talent_equipped(match) }
        return [
            self.create_summon('Large Wind Spirit', args)
        ] + super().get_actions(match)


# Talents


class ChaoticEntropy(SkillTalent):
    name: Literal['Chaotic Entropy']
    desc: str = (
        'Combat Action: When your active character is Sucrose, equip this '
        'card. After Sucrose equips this card, immediately use Forbidden '
        'Creation - Isomer 75 / Type II once. After the Large Wind Spirit '
        'created by your Sucrose, who had this card equipped, has converted '
        'to another Elemental Type: Deal +1 DMG of this Elemental Type.'
    )
    version: Literal['3.3'] = '3.3'
    charactor_name: Literal['Sucrose'] = 'Sucrose'
    cost: Cost = Cost(
        elemental_dice_color = DieColor.ANEMO,
        elemental_dice_number = 3,
        charge = 2
    )
    skill: ForbiddenCreationIsomer75TypeII = ForbiddenCreationIsomer75TypeII()


# charactor base


class Sucrose(CharactorBase):
    name: Literal['Sucrose']
    version: Literal['3.3'] = '3.3'
    desc: str = '''"Harmless Sweetie" Sucrose'''
    element: ElementType = ElementType.ANEMO
    max_hp: int = 10
    max_charge: int = 2
    skills: List[
        ElementalNormalAttackBase | AstableAnemohypostasisCreation6308 
        | ForbiddenCreationIsomer75TypeII
    ] = []
    faction: List[FactionType] = [
        FactionType.MONDSTADT
    ]
    weapon_type: WeaponType = WeaponType.CATALYST
    talent: ChaoticEntropy | None = None

    def _init_skills(self) -> None:
        self.skills = [
            ElementalNormalAttackBase(
                name = 'Wind Spirit Creation',
                damage_type = self.element,
                cost = ElementalNormalAttackBase.get_cost(self.element),
            ),
            AstableAnemohypostasisCreation6308(),
            ForbiddenCreationIsomer75TypeII()
        ]



import json
from typing import Literal, List
from fastapi import FastAPI, HTTPException
from fastapi.responses import JSONResponse
from fastapi.middleware.cors import CORSMiddleware
from pydantic import BaseModel
import uvicorn

from ..server.match import Match, MatchConfig
from ..server.deck import Deck
from ..agents import InteractionAgent
from .utils import get_new_match


class ResetData(BaseModel):
    fixed_random_seed: bool = False
    offset: int = 0
    match_config: MatchConfig | None = None
    rich_mode: bool = False
    match_state: Match | None = None
    match_state_idx: int | None = None


class DeckData(BaseModel):
    deck_str: str
    player_idx: int


class RespondData(BaseModel):
    player_idx: int
    command: str


class HTTPServer():
    """
    Simple HTTP server based on HTTP API. No safety protocol and data 
    visibility support now, i.e. everyone can get full information of opponent
    player and system states.

    To check APIs, use /docs or /redoc after run the server.
    """
    def __init__(
        self, 
        allow_origins: List[str] = ['*'], 
        decks: List[Deck | str] = [],
        match_config: MatchConfig | None = None,
    ):
        self.app = FastAPI()
        self.decks = [Deck.from_str(deck) if isinstance(deck, str) else deck
                      for deck in decks]
        if match_config is not None and match_config.history_level < 10:
            raise ValueError(
                'history_level must be at least 10 for HTTPServer')
        self.match: Match = get_new_match(
            self.decks, match_config = match_config)
        self.agent_0 = InteractionAgent(player_idx = 0, 
                                        only_use_command = True)
        self.agent_1 = InteractionAgent(player_idx = 1, 
                                        only_use_command = True)

        if len(self.decks) == 0:
            # no deck input at init, create two empty decks.
            self.decks = [Deck.from_str('''''') for _ in range(2)]

        # attrs to save history. command is saved separately for two players,
        # and each line contains [idx, command], idx is idx of last history.
        self.command_history = [[], []]
        self.match_random_state = self.match.random_state
        self.start_deck = self.decks[:]

        app = self.app
        # Add the CORSMiddleware to the app
        app.add_middleware(
            CORSMiddleware,
            allow_origins=allow_origins,
            # allow_origins=['*'],
            allow_credentials=True,
            allow_methods=['*'],
            allow_headers=['*'],
        )

        # @app.on_event('startup')
        # async def startup_event():
        #     self.match = get_new_match(
        #         decks = self.decks,
        #         rich_mode = False
        #     )

        '''API hanlders for app'''

        @app.post('/reset')
        async def reset(data: ResetData):
            """
            Reset the match. 
            If match_state_idx is not None, the match will be reset to the
            state of match_state_idx in history. This will keep histories 
            before match_state_idx.
            Otherwise, if match_state is not None, the match will be reset to
            match_state. This will clear all histories and start with zero.
            Otherwise, a new match will be created, with 
            fixed_random_seed, offset and match_config and rich_mode. Note when
            match_config is None, current match_config in self.match will be
            used.

            Return: The dict contains last state with its idx of the match if 
                success.
            """
            match = self.match
            fixed_random_seed = data.fixed_random_seed
            if fixed_random_seed:
                raise NotImplementedError('fixed_random_seed not supported')
            match_config = data.match_config
            if match_config is None:
                match_config = match.config
            rich = data.rich_mode
            match_state = data.match_state
            match_state_idx = data.match_state_idx
            if match_state_idx is not None:
                history = match._history[:]
                history_diff = match._history_diff[:]
                if len(history) <= match_state_idx or match_state_idx < 0:
                    raise HTTPException(status_code = 404, 
                                        detail = 'State not found')
                match = history[match_state_idx].copy(deep = True)
                match._history = history[:match_state_idx + 1]
                match._history_diff = history_diff[:match_state_idx + 1]
                self.match = match
                # remove command history that after match_state_idx. No need to
                # change match random state and start deck.
                current_cmd_hist = self.command_history[:]
                self.command_history = [[], []]
                for source, target in zip(current_cmd_hist,
                                          self.command_history):
                    for idx, cmd in source:
                        if idx >= match_state_idx:
                            break
                        target.append([idx, cmd])
            elif match_state is not None:
                match = match_state
                match._save_history()
                match_state_idx = len(match._history) - 1
                # when reset by match_state, cannot record history with only
                # match random state, set to None to notify that cannot save.
                self.command_history = [[], []]
                self.match_random_state = None
                self.start_deck = [x.player_deck_information 
                                   for x in match.player_tables]
            else:
                self.match = get_new_match(
                    decks = self.decks,
                    rich_mode = rich,
                    match_config = match_config,
                )
                match_state_idx = 0
                match = self.match
                self.command_history = [[], []]
                self.match_random_state = self.match.random_state
                self.start_deck = [x.player_deck_information 
                                   for x in match.player_tables]
            return {
                'idx': match_state_idx,
                'match': match.dict(),
                'type': 'FULL',
            }

        @app.get('/deck/{player_idx}')
        async def get_deck(player_idx: int):
            """
            get deck of a player.
            """
            if player_idx < 0 or player_idx > 1:
                raise HTTPException(status_code = 404, 
                                    detail = 'Player not found')
            return self.decks[player_idx].dict()

        @app.get('/decks')
        async def get_decks():
            """
            get decks of both players.
            """
            return [deck.dict() for deck in self.decks]

        @app.post('/deck')
        async def post_deck(data: DeckData):
            """
            Set deck.
            """
            match = self.match
            deck_str = data.deck_str
            player_idx = data.player_idx
            if player_idx < 0 or player_idx > 1:
                raise HTTPException(status_code = 404, 
                                    detail = 'Player not found')
            deck = Deck.from_str(deck_str)
            if not deck.check_legal(
                match.config.card_number, 
                match.config.max_same_card_number, 
                match.config.charactor_number, 
                match.config.check_deck_restriction
            ):
                raise HTTPException(status_code = 403, 
                                    detail = 'Deck not legal')
            self.decks[player_idx] = deck

        @app.get('/state/{mode}/{state_idx}/{player_idx}')
        async def get_game_state(
            mode: Literal['one', 'after'], state_idx: int, player_idx: int
        ):
            """
            Return list of state and its index.

            mode is one: only get the state of state_idx (but also wrapped in a
                list)
            mode is after: get the state of state_idx and all states after it

            state_idx: -1 means the last state; otherwise, it is the index of 
                the state

            player_idx is -1: fetch complete data
            player_idx is 0 or 1: fetch data for player idx (currently not 
                implemented)
            """
            match = self.match
            # player idx check
            if player_idx < -1 or player_idx > 1:
                raise HTTPException(status_code = 404, 
                                    detail = 'Player not found')
            if player_idx != -1:
                raise HTTPException(status_code = 404, 
                                    detail = 'player data fetch not suppoted')
            # state idx check
            if state_idx < -1 or state_idx > len(match._history):
                raise HTTPException(status_code = 404, 
                                    detail = 'State not found')
            if state_idx == -1:
                state_idx = len(match._history) - 1
            if state_idx == len(match._history):
                # ask for the state after the last state
                return JSONResponse([])
            result = [{
                'idx': state_idx,
                'match': match._history[state_idx].dict(),
                'type': 'FULL',
            }]
            if mode == 'after':
                result += [
                    {
                        'idx': state_idx + 1 + idx,
                        'match_diff': diff,
                        'type': 'DIFF',
                    } 
                    for idx, diff 
                    in enumerate(match._history_diff[state_idx + 1:])
                ]
            print('state output length', len(result))
            return JSONResponse(result)

        @app.get('/request/{player_idx}')
        async def get_request(player_idx: int):
            """
            Return the requests of player_idx. if player_idx is -1, return all 
            requests.
            """
            if player_idx < -1 or player_idx > 1:
                raise HTTPException(status_code = 404, 
                                    detail = 'Player not found')
            res = self.match.requests
            if player_idx != -1:
                res = [x for x in res if x.player_idx == player_idx]
            return JSONResponse([x.dict() for x in res])

        @app.post('/respond')
        async def post_respond(data: RespondData):
            match = self.match
            player_idx = data.player_idx
            command = data.command
            current_history_length = len(match._history)
            if not match.need_respond(player_idx):
                raise HTTPException(status_code = 404, 
                                    detail = 'Not your turn')
            if player_idx == 0:
                agent = self.agent_0
            elif player_idx == 1:
                agent = self.agent_1
            else:
                raise HTTPException(status_code = 404, 
                                    detail = 'Player not found')
            agent.commands = [command]
            resp = agent.generate_response(match)
            if resp is None:
                raise HTTPException(status_code = 404, 
                                    detail = 'Invalid command')
            match.respond(resp)
            match.step()
            for agent in (self.agent_0, self.agent_1):
                if (
                    agent.__class__ != InteractionAgent 
                    or len(agent.commands) > 0  # type: ignore
                ):
                    while match.need_respond(agent.player_idx):
                        resp = agent.generate_response(match)
                        assert resp is not None
                        match.respond(resp)
                        match.step()
            # after success respond, save command into command history
            self.command_history[player_idx].append([
                current_history_length - 1, command
            ])
            # generate response
            ret = []
            for idx, [state, diff] in enumerate(zip(
                    match._history[current_history_length:],
                    match._history_diff[current_history_length:],)):
                if idx == 0:
                    ret.append({
                        'idx': idx + current_history_length,
                        'match': state.dict(),
                        'type': 'FULL',
                    })
                else:
                    ret.append({
                        'idx': idx + current_history_length,
                        'match_diff': diff,
                        'type': 'DIFF',
                    })
            return JSONResponse(ret)

        @app.get('/log')
        def get_log():
            """
            Return the log of the match. It is encoded in json.
            """
            try:
                log = self.save_log()
                return json.loads(log)
            except Exception as e:
                raise HTTPException(status_code = 500, detail = str(e))

    def run(self, *argv, **kwargs):
        """
        A wrapper of uvicorn.run
        """
        if len(argv):
            raise ValueError('positional arguments not supported')
        uvicorn.run(self.app, **kwargs)

    def save_log(self, file_path: str | None = None) -> str:
        """
        Save the log to specified file and return log str.
        If file path not specified, only return log str.
        log str is encoded in json.

        log format:
            match_random_state: initial random state of match
            start_deck: start deck string of players
            command_history: command history of players, different from
                self.command_history, it do not contain idx of history.
        """
        if self.match_random_state is None:
            raise RuntimeError('Cannot save log when match random state is '
                               'None. This may caused by reset match by '
                               'match_state.')
        data = {
            'match_random_state': self.match_random_state,
            'start_deck': [x.to_str() for x in self.start_deck],
            'command_history': [[y[1] for y in x] 
                                for x in self.command_history],
        }
        data_str = json.dumps(data)
        if file_path is not None:
            with open(file_path, 'w') as f:
                f.write(data_str)
        return data_str



from typing import Any, List, Literal

from ...event import GameStartEventArguments

from ...action import Actions, ChangeObjectUsageAction, CreateObjectAction
from ...struct import Cost

from ...consts import (
    DamageElementalType, DieColor, ElementType, FactionType, WeaponType
)
from ..charactor_base import (
    ElementalBurstBase, ElementalSkillBase, 
    PhysicalNormalAttackBase, PassiveSkillBase, CharactorBase, SkillTalent
)


# Charactor status. DO NOT define here, define in server/status/characor_status
# Here is just example.


# Round status, will last for several rounds and disappear
# Skills


class SecretRiteChasmicSoulfarer(ElementalSkillBase):
    name: Literal[
        'Secret Rite: Chasmic Soulfarer'] = 'Secret Rite: Chasmic Soulfarer'
    damage: int = 3
    damage_type: DamageElementalType = DamageElementalType.ELECTRO
    cost: Cost = Cost(
        elemental_dice_color = DieColor.ELECTRO,
        elemental_dice_number = 3
    )

    def get_actions(self, match: Any) -> List[Actions]:
        """
        When talent equipped and level match, increase damage by 1
        """
        if self.is_talent_equipped(match):
            status = match.player_tables[self.position.player_idx].charactors[
                self.position.charactor_idx].status
            for s in status:
                if s.name == 'Pactsworn Pathclearer':  # pragma: no branch
                    if s.usage == 3 or s.usage == 5:
                        self.damage = 4
                    break
            else:
                raise AssertionError('No Pactsworn Pathclearer status')
        ret = super().get_actions(match)
        self.damage = 3
        return ret


class SacredRiteWolfsSwiftness(ElementalBurstBase):
    name: Literal[
        "Sacred Rite: Wolf's Swiftness"] = "Sacred Rite: Wolf's Swiftness"
    desc: str = (
        "Deals 4 Electro DMG. Pactsworn Pathclearer's Indwelling Level +2."
    )
    damage: int = 4
    damage_type: DamageElementalType = DamageElementalType.ELECTRO
    cost: Cost = Cost(
        elemental_dice_color = DieColor.ELECTRO,
        elemental_dice_number = 4,
        charge = 2
    )

    def get_actions(self, match: Any) -> List[Actions]:
        """
        attack and increase usage
        """
        ret = super().get_actions(match)
        status = match.player_tables[self.position.player_idx].charactors[
            self.position.charactor_idx].status
        for s in status:
            if s.name == 'Pactsworn Pathclearer':  # pragma: no branch
                ret.append(ChangeObjectUsageAction(
                    object_position = s.position,
                    change_type = 'DELTA',
                    change_usage = 2
                ))
                return ret
        else:
            raise AssertionError('No Pactsworn Pathclearer status')


class LawfulEnforcer(PassiveSkillBase):
    name: Literal['Lawful Enforcer'] = 'Lawful Enforcer'
    desc: str = (
        '(Passive) When the battle begins, this character gains Pactsworn '
        'Pathclearer.'
    )

    def event_handler_GAME_START(
        self, event: GameStartEventArguments, match: Any
    ) -> List[CreateObjectAction]:
        """
        When game begin, gain Pactsworn Pathclearer
        """
        return [self.create_charactor_status('Pactsworn Pathclearer')]


# Talents


class FeatherfallJudgment(SkillTalent):
    name: Literal['Featherfall Judgment']
    desc: str = (
        'Combat Action: When your active character is Cyno, equip this card. '
        'After Cyno equips this card, immediately use Secret Rite: Chasmic '
        'Soulfarer once. When your Cyno, who has this card equipped, uses '
        "Secret Rite: Chasmic Soulfarer with 3 or 5 levels of Pactsworn "
        "Pathclearer's Indwelling effect, deal +1 additional DMG."
    )
    version: Literal['3.3'] = '3.3'
    charactor_name: Literal['Cyno'] = 'Cyno'
    cost: Cost = Cost(
        elemental_dice_color = DieColor.ELECTRO,
        elemental_dice_number = 3,
    )
    skill: SecretRiteChasmicSoulfarer = SecretRiteChasmicSoulfarer()


# charactor base


class Cyno(CharactorBase):
    name: Literal['Cyno']
    version: Literal['3.3'] = '3.3'
    desc: str = (
        '''"Judicator of Secrets" Cyno'''
    )
    element: ElementType = ElementType.ELECTRO
    max_hp: int = 10
    max_charge: int = 2
    skills: List[
        PhysicalNormalAttackBase | SecretRiteChasmicSoulfarer
        | SacredRiteWolfsSwiftness | LawfulEnforcer
    ] = []
    faction: List[FactionType] = [
        FactionType.SUMERU
    ]
    weapon_type: WeaponType = WeaponType.POLEARM
    talent: FeatherfallJudgment | None = None

    def _init_skills(self) -> None:
        self.skills = [
            PhysicalNormalAttackBase(
                name = "Invoker's Spear",
                cost = PhysicalNormalAttackBase.get_cost(self.element),
            ),
            SecretRiteChasmicSoulfarer(),
            SacredRiteWolfsSwiftness(),
            LawfulEnforcer()
        ]



from typing import Any, List, Literal

from ...action import Actions

from ...struct import Cost

from ...consts import (
    DamageElementalType, DieColor, ElementType, FactionType, WeaponType
)
from ..charactor_base import (
    ElementalBurstBase, ElementalSkillBase, 
    PhysicalNormalAttackBase, CharactorBase, SkillTalent
)


# Skills


class BaneOfAllEvil(ElementalBurstBase):
    name: Literal['Bane of All Evil'] = 'Bane of All Evil'
    desc: str = '''Deals 4 Anemo DMG. This character gains Yaksha's Mask.'''
    damage: int = 4
    damage_type: DamageElementalType = DamageElementalType.ANEMO
    cost: Cost = Cost(
        elemental_dice_color = DieColor.ANEMO,
        elemental_dice_number = 3,
        charge = 2
    )

    def get_actions(self, match: Any) -> List[Actions]:
        """
        If has talent, add skill cost decrease by 2
        """
        args = {}
        if self.is_talent_equipped(match):
            args['skill_cost_decrease_usage'] = 2
        return super().get_actions(match) + [
            self.create_charactor_status("Yaksha's Mask", args),
        ]


# Talents


class ConquerorOfEvilGuardianYaksha(SkillTalent):
    name: Literal['Conqueror of Evil: Guardian Yaksha']
    desc: str = (
        'Combat Action: When your active character is Xiao, equip this card. '
        'After Xiao equips this card, immediately use Bane of All Evil once. '
        "While your Xiao has Yaksha's Mask attached, your use of Lemniscatic "
        "Wind Cycling will cost 1 less Genius Invokation TCG Anemo Cost Anemo "
        "Die. (Every attachment of Yaksha's Mask allows the effect to be "
        "triggered twice)"
    )
    version: Literal['3.7'] = '3.7'
    charactor_name: Literal['Xiao'] = 'Xiao'
    cost: Cost = Cost(
        elemental_dice_color = DieColor.ANEMO,
        elemental_dice_number = 3,
        charge = 2
    )
    skill: BaneOfAllEvil = BaneOfAllEvil()


# charactor base


class Xiao(CharactorBase):
    name: Literal['Xiao']
    version: Literal['3.7'] = '3.7'
    desc: str = '''"Vigilant Yaksha" Xiao'''
    element: ElementType = ElementType.ANEMO
    max_hp: int = 10
    max_charge: int = 2
    skills: List[
        PhysicalNormalAttackBase | ElementalSkillBase | BaneOfAllEvil
    ] = []
    faction: List[FactionType] = [
        FactionType.LIYUE
    ]
    weapon_type: WeaponType = WeaponType.POLEARM
    talent: ConquerorOfEvilGuardianYaksha | None = None

    def _init_skills(self) -> None:
        self.skills = [
            PhysicalNormalAttackBase(
                name = 'Whirlwind Thrust',
                cost = PhysicalNormalAttackBase.get_cost(self.element),
            ),
            ElementalSkillBase(
                name = 'Lemniscatic Wind Cycling',
                damage_type = DamageElementalType.ANEMO,
                cost = ElementalSkillBase.get_cost(self.element),
            ),
            BaneOfAllEvil(),
        ]



import pytest

from src.lpsim.agents.random_agent import RandomAgent
from src.lpsim.agents.nothing_agent import NothingAgent
from src.lpsim.server.action import CreateDiceAction, MoveObjectAction
from src.lpsim.server.consts import ObjectPositionType
from src.lpsim.server.deck import Deck
from src.lpsim.server.interaction import SwitchCardResponse
from src.lpsim.server.object_base import ObjectBase
from src.lpsim.server.struct import ObjectPosition
from src.lpsim.server.match import Match, MatchConfig


def test_object_position_validation():
    p1 = ObjectPosition(
        player_idx = 0,
        charactor_idx = 1,
        area = ObjectPositionType.CHARACTOR_STATUS,
        id = 345,
    )
    p2 = ObjectPosition(
        player_idx = 1,
        charactor_idx = 0,
        area = ObjectPositionType.CHARACTOR,
        id = 123
    )
    match = Match()
    match.player_tables[0].active_charactor_idx = 1
    match.player_tables[1].active_charactor_idx = 0
    assert p1.check_position_valid(p2, match)
    assert not p1.check_position_valid(p2, match, player_idx_same = True)
    assert not p1.check_position_valid(p2, match, charactor_idx_same = True)
    assert not p1.check_position_valid(p2, match, player_idx_same = True, 
                                       charactor_idx_same = True)
    assert not p1.check_position_valid(p2, match, area_same = True)
    assert not p1.check_position_valid(p2, match, id_same = True)
    assert p1.check_position_valid(p2, match, area_same = False)
    assert p1.check_position_valid(p2, match, id_same = False)
    assert p1.check_position_valid(
        p2, match, source_area = ObjectPositionType.CHARACTOR_STATUS)
    assert p1.check_position_valid(
        p2, match, target_area = ObjectPositionType.CHARACTOR)
    assert p1.check_position_valid(
        p2, match, source_is_active_charactor = True)
    assert p1.check_position_valid(
        p2, match, target_is_active_charactor = True)
    assert not p1.check_position_valid(
        p2, match, source_area = ObjectPositionType.CHARACTOR)
    assert not p1.check_position_valid(
        p2, match, target_area = ObjectPositionType.CHARACTOR_STATUS)
    assert not p1.check_position_valid(
        p2, match, source_is_active_charactor = False)
    assert not p1.check_position_valid(
        p2, match, target_is_active_charactor = False)
    assert p1.check_position_valid(
        p2, match, player_idx_same = False)
    assert p1.check_position_valid(
        p2, match, player_idx_same = False, charactor_idx_same = False)
    assert not p1.check_position_valid(
        p2, match, player_idx_same = True, charactor_idx_same = False)
    assert not p1.check_position_valid(
        p2, match, player_idx_same = False, charactor_idx_same = True)


def test_objectbase_wrong_handler_name():
    class WrongEventHandler(ObjectBase):
        position: ObjectPosition = ObjectPosition(
            player_idx = -1,
            area = ObjectPositionType.INVALID,
            id = -1,
        )

        def event_handler_NOT_EXIST(self, event):
            ...  # pragma: no cover

    with pytest.raises(ValueError):
        _ = WrongEventHandler()

    class WrongValueModifier(ObjectBase):
        position: ObjectPosition = ObjectPosition(
            player_idx = -1,
            area = ObjectPositionType.INVALID,
            id = -1,
        )

        def value_modifier_NOT_EXIST(self, value, match, mode):
            ...  # pragma: no cover

    with pytest.raises(ValueError):
        _ = WrongValueModifier()


def test_match_config_and_match_errors():
    config = MatchConfig()
    assert config.check_config()
    config.initial_hand_size = 11
    assert not config.check_config()
    config.initial_hand_size = 5
    config.initial_card_draw = 11
    assert not config.check_config()
    config.max_hand_size = 999
    config.card_number = 10
    assert not config.check_config()
    config.initial_card_draw = 2
    config.initial_dice_number = 20
    assert not config.check_config()
    config.initial_dice_number = 8
    config = MatchConfig(initial_hand_size = -1)
    assert not config.check_config()
    config = MatchConfig(initial_card_draw = -1)
    assert not config.check_config()
    config = MatchConfig(initial_dice_number = -1)
    assert not config.check_config()
    config = MatchConfig(initial_dice_reroll_times = -1)
    assert not config.check_config()
    config = MatchConfig(card_number = -1)
    assert not config.check_config()
    config = MatchConfig(max_same_card_number = -1)
    assert not config.check_config()
    config = MatchConfig(charactor_number = -1)
    assert not config.check_config()
    config = MatchConfig(max_round_number = 0)
    assert not config.check_config()
    config = MatchConfig(max_round_number = -1)
    assert not config.check_config()
    config = MatchConfig(max_hand_size = -1)
    assert not config.check_config()
    config = MatchConfig(max_dice_number = -1)
    assert not config.check_config()
    config = MatchConfig(max_summon_number = -1)
    assert not config.check_config()
    config = MatchConfig(max_support_number = -1)
    assert not config.check_config()
    config = MatchConfig(history_level = -1)
    assert not config.check_config()
    match = Match()
    match.config = config
    assert not match.start()
    match = Match()
    match.player_tables = []
    assert not match.start()
    match = Match()
    deck = Deck.from_str(
        '''
        default_version:4.0
        charactor:GeoMob*3
        Strategize*100
        '''
    )
    with pytest.raises(AssertionError):
        match.set_deck([])
    match.set_deck([deck, deck])
    assert not match.start()
    deck = Deck.from_str(
        '''
        default_version:4.0
        charactor:GeoMob*3
        Strategize*30
        '''
    )
    match = Match()
    match.config.max_same_card_number = 30
    match.set_deck([deck, deck])
    assert match.start()
    agent_0 = NothingAgent(player_idx = 0)
    agent_1 = RandomAgent(player_idx = 1)
    non_exist_tested = False
    respond_no_request = False
    req = None
    resp = None
    for _ in range(30):
        if (
            len(match.requests) == 0 
            and not respond_no_request 
            and non_exist_tested
        ):
            assert not match.need_respond(0)
            assert not match.need_respond(1)
            assert resp is not None
            with pytest.raises(ValueError):
                match.respond(resp)
        if match.need_respond(0):
            if not non_exist_tested:
                assert match.need_respond(1)
                non_exist_tested = True
                req = match.requests[0].copy(deep = True)
                assert match.check_request_exist(req)
                resp = SwitchCardResponse(
                    request = req, card_idxs = [-1])  # type: ignore
                with pytest.raises(ValueError):
                    match.respond(resp)
                resp.card_idxs = []
                resp.request.player_idx = 2
                with pytest.raises(ValueError):
                    match.respond(resp)
            resp = agent_0.generate_response(match)
            assert resp is not None
            match.respond(resp)
        elif match.need_respond(1):
            resp = agent_1.generate_response(match)
            assert resp is not None
            match.respond(resp)
        if len(match.requests) == 0:
            match.step(run_continuously = False)


def test_create_dice():
    match = Match()
    act = CreateDiceAction(
        player_idx = 0,
        number = 4,
        random = True,
        different = True
    )
    with pytest.raises(ValueError):
        match._act(act)
    act = CreateDiceAction(
        player_idx = 0,
        number = 4,
    )
    with pytest.raises(ValueError):
        match._act(act)
    act = CreateDiceAction(
        player_idx = 0,
        number = 8,
        different = True,
    )
    with pytest.raises(ValueError):
        # different over maximum color number
        match._act(act)
    act = CreateDiceAction(
        player_idx = 0,
        number = 7,
        different = True,
    )
    match._act(act)
    assert len(match.player_tables[0].dice.colors) == 7
    assert len(set(match.player_tables[0].dice.colors)) == 7
    match.player_tables[0].dice.colors.clear()
    for i in range(50):
        act = CreateDiceAction(
            player_idx = 0,
            number = 3,
            different = True,
        )
        match._act(act)
        assert len(match.player_tables[0].dice.colors) == 3
        assert len(set(match.player_tables[0].dice.colors)) == 3
        for color in match.player_tables[0].dice.colors:
            assert color != 'OMNI'
        match.player_tables[0].dice.colors.clear()


def test_id_wont_duplicate():
    s = set()
    position = ObjectPosition(
        player_idx = -1,
        area = ObjectPositionType.INVALID,
        id = -1,
    )
    for _ in range(50000):
        o = ObjectBase(
            position = position,
        )
        assert o.id not in s
        s.add(o.id)


def test_remove_non_exist_equip():
    match = Match()
    deck = Deck.from_str(
        '''
        default_version:4.0
        charactor:GeoMob*3
        Strategize*30
        '''
    )
    match.set_deck([deck, deck])
    match.config.max_same_card_number = None
    match.config.charactor_number = None
    match.config.card_number = None
    match.config.check_deck_restriction = False
    assert match.start()
    with pytest.raises(AssertionError):
        match._action_move_object(
            MoveObjectAction(
                object_position = ObjectPosition(
                    player_idx = 0,
                    charactor_idx = 0,
                    area = ObjectPositionType.CHARACTOR,
                    id = 0,
                ),
                target_position = ObjectPosition(
                    player_idx = 0,
                    area = ObjectPositionType.HAND,
                    id = 0,
                ),
            )
        )


def test_immutable_position():
    position = ObjectPosition(
        player_idx = 0,
        area = ObjectPositionType.SUMMON,
        id = 0
    )
    with pytest.raises(AttributeError):
        position.player_idx = 1
    with pytest.raises(AttributeError):
        del position.area


if __name__ == '__main__':
    # test_object_position_validation()
    # test_match_config_and_match_errors()
    # test_objectbase_wrong_handler_name()
    # test_create_dice()
    # test_id_wont_duplicate()
    test_remove_non_exist_equip()



from typing import Any, List, Literal

from ...struct import Cost

from ...action import (
    Actions, ChangeObjectUsageAction, MakeDamageAction, RemoveObjectAction, 
    SkillEndAction
)

from ...event import (
    ChangeObjectUsageEventArguments, MakeDamageEventArguments, 
    RoundEndEventArguments, SkillEndEventArguments
)

from ...consts import (
    DamageElementalType, DamageType, DieColor, ElementType, IconType, 
    ObjectPositionType, SkillType
)

from ...modifiable_values import (
    CostValue, DamageElementEnhanceValue, DamageIncreaseValue, DamageValue
)
from .base import (
    ElementalInfusionCharactorStatus, PrepareCharactorStatus, 
    ReviveCharactorStatus, RoundCharactorStatus, ShieldCharactorStatus, 
    UsageCharactorStatus, CharactorStatusBase
)


class ElectroInfusionKeqing(ElementalInfusionCharactorStatus,
                            RoundCharactorStatus):
    """
    Inherit from vanilla elemental infusion status, and has talent activated
    argument. when talent activated, it will gain electro damage +1 buff,
    and usage +1.
    """
    name: Literal['Electro Elemental Infusion'] = 'Electro Elemental Infusion'
    mark: Literal['Keqing']  # used to select right status
    buff_desc: str = (
        'When the charactor to which it is attached to deals '
        'Physical Damage, it will be turned into Electro DMG, '
        'and Electro DMG dealt by the charactor +1.'
    )
    version: Literal['3.3'] = '3.3'
    usage: int = 2
    max_usage: int = 2

    talent_activated: bool = False

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if self.talent_activated:
            self.desc = self.buff_desc

    def renew(self, new_status: 'ElectroInfusionKeqing') -> None:
        super().renew(new_status)
        if new_status.talent_activated:
            self.talent_activated = True
            self.desc = self.buff_desc

    def value_modifier_DAMAGE_INCREASE(
        self, value: DamageIncreaseValue, match: Any,
        mode: Literal['TEST', 'REAL']
    ) -> DamageIncreaseValue:
        """
        If talent activated, increase electro damage dealt by this charactor 
        by 1.
        """
        if not self.talent_activated:
            # talent not activated, do nothing
            return value
        if not value.is_corresponding_charactor_use_damage_skill(
            self.position, match, None
        ):
            # not corresponding charactor use skill, do nothing
            return value
        if value.damage_elemental_type != DamageElementalType.ELECTRO:
            # not electro damage, do nothing
            return value
        value.damage += 1
        return value


class ElectroCrystalCore(ReviveCharactorStatus):
    name: Literal['Electro Crystal Core'] = 'Electro Crystal Core'
    version: Literal['3.7'] = '3.7'
    heal: int = 1


class RockPaperScissorsComboScissors(PrepareCharactorStatus):
    name: Literal[
        'Rock-Paper-Scissors Combo: Scissors'
    ] = 'Rock-Paper-Scissors Combo: Scissors'
    desc: str = (
        'Prepare Skill Rock-Paper-Scissors Combo: Scissors and '
        'Rock-Paper-Scissors Combo: Paper.'
    )
    version: Literal['3.7'] = '3.7'
    charactor_name: Literal['Electro Hypostasis'] = 'Electro Hypostasis'
    skill_name: Literal[
        'Rock-Paper-Scissors Combo: Scissors'
    ] = 'Rock-Paper-Scissors Combo: Scissors'


class RockPaperScissorsComboPaper(PrepareCharactorStatus):
    name: Literal[
        'Rock-Paper-Scissors Combo: Paper'
    ] = 'Rock-Paper-Scissors Combo: Paper'
    desc: str = (
        'Prepare Skill Rock-Paper-Scissors Combo: Paper.'
    )
    version: Literal['3.7'] = '3.7'
    charactor_name: Literal['Electro Hypostasis'] = 'Electro Hypostasis'
    skill_name: Literal[
        'Rock-Paper-Scissors Combo: Paper'
    ] = 'Rock-Paper-Scissors Combo: Paper'


class ChakraDesiderata(CharactorStatusBase):
    name: Literal['Chakra Desiderata'] = 'Chakra Desiderata'
    desc: str = (
        'After your other characters use Elemental Bursts: Gain 1 Resolve. '
        '(Max 3) '
        'When the character to which this is attached uses Secret Art: Musou '
        'Shinsetsu: Consume all Resolve and deal +1 DMG per Resolve.'
    )
    version: Literal['3.7'] = '3.7'
    usage: int = 0
    max_usage: int = 3
    icon_type: Literal[IconType.OTHERS] = IconType.OTHERS

    def event_handler_SKILL_END(
        self, event: SkillEndEventArguments, match: Any
    ) -> List[Actions]:
        """
        When other charactor using elemental burst, gain resolve.
        """
        if not self.position.check_position_valid(
            event.action.position, match, player_idx_same = True,
            charactor_idx_same = False, target_area = ObjectPositionType.SKILL,
        ):
            # not other ally use elemental burst, do nothing
            return []
        if event.action.skill_type == SkillType.ELEMENTAL_BURST:
            # is burst
            self.usage = min(self.usage + 1, self.max_usage)
        return []

    def value_modifier_DAMAGE_INCREASE(
        self, value: DamageIncreaseValue, match: Any,
        mode: Literal['TEST', 'REAL']
    ) -> DamageIncreaseValue:
        """
        When self using elemental burst, increase damage by usage.
        """
        if not value.is_corresponding_charactor_use_damage_skill(
            self.position, match, SkillType.ELEMENTAL_BURST
        ):
            # not self use burst, do nothing
            return value
        # increase damage
        value.damage += self.usage
        charactor = match.player_tables[self.position.player_idx].charactors[
            self.position.charactor_idx]
        if charactor.talent is not None:
            # has talent, double increase damage
            value.damage += self.usage
        self.usage = 0
        return value


class TheShrinesSacredShade(RoundCharactorStatus):
    name: Literal["The Shrine's Sacred Shade"] = "The Shrine's Sacred Shade"
    desc: str = (
        'During this round, the next Yakan Evocation: Sesshou Sakura used by '
        'the charactor to which this is attached will cost 2 less Elemental '
        'Dice.'
    )
    version: Literal['3.7'] = '3.7'
    usage: int = 1
    max_usage: int = 1
    icon_type: Literal[IconType.SPECIAL] = IconType.SPECIAL

    def value_modifier_COST(
        self, value: CostValue, match: Any, mode: Literal['TEST', 'REAL'],
    ) -> CostValue:
        """
        When use Yakan Evocation: Sesshou Sakura, cost 2 less
        """
        if not self.position.check_position_valid(
            value.position, match, player_idx_same = True,
            charactor_idx_same = True,
            target_area = ObjectPositionType.SKILL
        ):
            # not our player skill, do nothing
            return value
        skill = match.get_object(value.position)
        if skill.name != 'Yakan Evocation: Sesshou Sakura':
            # not right skill, do nothing
            return value
        assert self.usage > 0
        success = value.cost.decrease_cost(DieColor.ELECTRO)
        success = value.cost.decrease_cost(DieColor.ELECTRO) or success
        if success and mode == 'REAL':
            self.usage -= 1
        return value

    def event_handler_SKILL_END(
        self, event: SkillEndAction, match: Any
    ) -> List[RemoveObjectAction]:
        return self.check_should_remove()


class TheWolfWithin(RoundCharactorStatus):
    name: Literal['The Wolf Within'] = 'The Wolf Within'
    desc: str = (
        'After the character to which this is attached uses a Normal Attack '
        'or an Elemental Skill: Deal 2 Electro DMG.'
    )
    version: Literal['3.3'] = '3.3'
    usage: int = 2
    max_usage: int = 2
    icon_type: Literal[IconType.OTHERS] = IconType.OTHERS

    def event_handler_SKILL_END(
        self, event: SkillEndEventArguments, match: Any
    ) -> List[MakeDamageAction]:
        """
        If self use normal attack or elemental skill, deal damage
        """
        if not self.position.check_position_valid(
            event.action.position, match, player_idx_same = True,
            charactor_idx_same = True, target_area = ObjectPositionType.SKILL
        ):
            # not self use skill
            return []
        if event.action.skill_type not in [
            SkillType.NORMAL_ATTACK, SkillType.ELEMENTAL_SKILL
        ]:
            # not normal attack or elemental skill
            return []
        # make damage
        target = match.player_tables[
            1 - self.position.player_idx].get_active_charactor()
        return [MakeDamageAction(
            source_player_idx = self.position.player_idx,
            target_player_idx = 1 - self.position.player_idx,
            damage_value_list = [
                DamageValue(
                    position = self.position,
                    damage_type = DamageType.DAMAGE,
                    target_position = target.position,
                    damage = 2,
                    damage_elemental_type = DamageElementalType.ELECTRO,
                    cost = Cost()
                )
            ]
        )]


class TidecallerSurfEmbrace(ShieldCharactorStatus, PrepareCharactorStatus):
    name: Literal['Tidecaller: Surf Embrace'] = 'Tidecaller: Surf Embrace'
    desc: str = (
        'The next time this character acts, they will immediately use the '
        'Skill Wavestrider. While preparing this Skill: Grant 2 Shield '
        'points to the character to which this is attached.'
    )
    version: Literal['3.4'] = '3.4'
    charactor_name: Literal['Beidou'] = 'Beidou'
    skill_name: Literal['Wavestrider'] = 'Wavestrider'

    usage: int = 2
    max_usage: int = 2

    def event_handler_MAKE_DAMAGE(
        self, event: MakeDamageEventArguments, match: Any
    ) -> List[Actions]:
        """
        Do not remove when usage becomes zero
        """
        return []


class CrowfeatherCover(UsageCharactorStatus):
    name: Literal['Crowfeather Cover'] = 'Crowfeather Cover'
    desc: str = (
        'The character with this attached deals +1 Elemental Skill and '
        'Elemental Burst DMG.'
    )
    version: Literal['3.5'] = '3.5'
    usage: int = 2
    max_usage: int = 2
    icon_type: Literal[IconType.ATK_UP] = IconType.ATK_UP

    def value_modifier_DAMAGE_INCREASE(
        self, value: DamageIncreaseValue, match: Any,
        mode: Literal['TEST', 'REAL'],
    ) -> DamageIncreaseValue:
        if value.damage_from_element_reaction:
            # is elemental reaction, do nothing
            return value
        if (
            value.is_corresponding_charactor_use_damage_skill(
                self.position, match, SkillType.ELEMENTAL_SKILL
            )
            or value.is_corresponding_charactor_use_damage_skill(
                self.position, match, SkillType.ELEMENTAL_BURST
            )
        ):
            # is self use elemental skill or elemental burst
            if self.usage <= 0:  # pragma: no cover
                # no usage
                return value
            # increase damage
            assert mode == 'REAL'
            self.usage -= 1
            value.damage += 1
            # talent effects
            charactors = match.player_tables[
                self.position.player_idx].charactors
            if (
                charactors[self.position.charactor_idx].element 
                != ElementType.ELECTRO
            ):
                # not electro
                return value
            found_talent_kujou: bool = False
            for charactor in charactors:
                if (
                    charactor.name == 'Kujou Sara' 
                    and charactor.talent is not None
                ):
                    found_talent_kujou = True
                    break
            if found_talent_kujou:
                # increase one more damage
                value.damage += 1
        return value


class PactswornPathclearer(ElementalInfusionCharactorStatus):
    name: Literal['Pactsworn Pathclearer'] = 'Pactsworn Pathclearer'
    desc: str = (
        'End Phase: Gain 1 level of Indwelling. '
        'Create the following effects based on the Indwelling Level: '
        'If equal to or greater than 2: Physical DMG dealt is converted to '
        'Electro DMG; If equal or greater than 4: DMG dealt +2; If equal to '
        'or greater than 6: Indwelling Level -4'
    )
    version: Literal['3.3'] = '3.3'
    usage: int = 0
    max_usage: int = 999
    infused_elemental_type: DamageElementalType = DamageElementalType.ELECTRO
    icon_type: Literal[IconType.OTHERS] = IconType.OTHERS

    def value_modifier_DAMAGE_ELEMENT_ENHANCE(
        self, value: DamageElementEnhanceValue, match: Any, 
        mode: Literal['TEST', 'REAL']
    ) -> DamageElementEnhanceValue:
        if self.usage < 2:
            # level not enough
            return value
        return super().value_modifier_DAMAGE_ELEMENT_ENHANCE(
            value, match, mode)

    def value_modifier_DAMAGE_INCREASE(
        self, value: DamageIncreaseValue, match: Any,
        mode: Literal['TEST', 'REAL'],
    ) -> DamageIncreaseValue:
        if self.usage < 4:
            # level not enough
            return value
        if value.is_corresponding_charactor_use_damage_skill(
            self.position, match, None
        ):
            # self use skill, increase damage
            assert mode == 'REAL'
            value.damage += 2
        return value

    def event_handler_ROUND_END(
        self, event: RoundEndEventArguments, match: Any
    ) -> List[ChangeObjectUsageAction]:
        """
        add one usage
        """
        return [ChangeObjectUsageAction(
            object_position = self.position,
            change_type = 'DELTA',
            change_usage = 1
        )]

    def event_handler_CHANGE_OBJECT_USAGE(
        self, event: ChangeObjectUsageEventArguments, match: Any
    ) -> List[Actions]:
        """
        When change object usage, check if usage is greater than 6.
        If so, decrease by 4.
        """
        if self.usage >= 6:
            self.usage -= 4
        return []


class Conductive(CharactorStatusBase):
    """
    repeatedly attach, end phase accumulate will perform in this status;
    damage increase will perform in elemental skill.
    """
    name: Literal['Conductive'] = 'Conductive'
    desc: str = (
        'This status starts with 2 stacks of Conductive. When attached '
        'repeatedly, Conductive stack +1. Conductive can be stacked to a '
        'maximum of 4 stacks. End Phase: Accumulate 1 stack of Conductive. '
        'When the character attached with Conductive takes DMG from Violet '
        'Arc: Remove this status and DMG +1 for each stack of Conductive. '
    )
    version: Literal['4.0'] = '4.0'
    usage: int = 2
    max_usage: int = 4
    icon_type: Literal[IconType.DEBUFF] = IconType.DEBUFF

    def renew(self, new_status: 'Conductive'):
        """
        Add one usage
        """
        self.usage = min(self.max_usage, self.usage + 1)

    def event_handler_ROUND_END(
        self, event: RoundEndEventArguments, match: Any
    ) -> List[Actions]:
        """
        When in round end, add one usage
        """
        self.usage = min(self.max_usage, self.usage + 1)
        return []


ElectroCharactorStatus = (
    ElectroInfusionKeqing | RockPaperScissorsComboScissors
    | RockPaperScissorsComboPaper | ElectroCrystalCore | ChakraDesiderata
    | TheShrinesSacredShade | TheWolfWithin | TidecallerSurfEmbrace
    | CrowfeatherCover | PactswornPathclearer | Conductive
)



from typing import Any, List, Literal

from ...summon.base import DefendSummonBase

from ...modifiable_values import DamageValue

from ...action import Actions, MakeDamageAction, RemoveObjectAction
from ...struct import Cost

from ...consts import (
    DamageElementalType, DamageType, DieColor, ElementType, FactionType, 
    WeaponType
)
from ..charactor_base import (
    AOESkillBase, ElementalBurstBase, ElementalSkillBase, 
    PhysicalNormalAttackBase, CharactorBase, SkillTalent
)


# Summons


class BaronBunny(DefendSummonBase):
    name: Literal['Baron Bunny'] = 'Baron Bunny'
    desc: str = (
        'When your active character takes DMG: Decrease DMG by 2. '
        'When the Usages are depleted, this card will not be discarded. '
        'At the End Phase, if Usage(s) have been depleted: Discard this card, '
        'deal 2 Pyro DMG.'
    )
    version: Literal['3.7'] = '3.7'
    usage: int = 1
    max_usage: int = 1
    damage_elemental_type: DamageElementalType = DamageElementalType.PYRO
    damage: int = 2
    min_damage_to_trigger: int = 1
    max_in_one_time: int = 2
    attack_until_run_out_of_usage: bool = True


# Skills


class Sharpshooter(PhysicalNormalAttackBase):
    name: Literal['Sharpshooter'] = 'Sharpshooter'
    cost: Cost = PhysicalNormalAttackBase.get_cost(ElementType.PYRO)

    def get_actions(self, match: Any) -> List[Actions]:
        """
        If has talent, and Baron Bunny in area, do extra 3 pyro damage and 
        remove Baron Bunny.
        """
        ret = super().get_actions(match)
        if self.is_talent_equipped(match):
            summons = match.player_tables[self.position.player_idx].summons
            for summon in summons:
                if summon.name == 'Baron Bunny':
                    target = match.player_tables[
                        1 - self.position.player_idx].get_active_charactor()
                    ret.append(MakeDamageAction(
                        source_player_idx = self.position.player_idx,
                        target_player_idx = 1 - self.position.player_idx,
                        damage_value_list = [
                            DamageValue(
                                position = summon.position,
                                damage_type = DamageType.DAMAGE,
                                target_position = target.position,
                                damage = 3,
                                damage_elemental_type
                                = DamageElementalType.PYRO,
                                cost = Cost(),
                            )
                        ]
                    ))
                    ret.append(RemoveObjectAction(
                        object_position = summon.position
                    ))
                    return ret
        return ret


class ExplosivePuppet(ElementalSkillBase):
    name: Literal['Explosive Puppet'] = 'Explosive Puppet'
    desc: str = '''Summons 1 Baron Bunny.'''
    damage: int = 0
    damage_type: DamageElementalType = DamageElementalType.PIERCING
    cost: Cost = Cost(
        elemental_dice_color = DieColor.PYRO,
        elemental_dice_number = 3
    )

    def get_actions(self, match: Any) -> List[Actions]:
        """
        create object only
        """
        return [
            self.charge_self(1),
            self.create_summon('Baron Bunny'),
        ]


class FieryRain(ElementalBurstBase, AOESkillBase):
    name: Literal['Fiery Rain'] = 'Fiery Rain'
    damage: int = 2
    damage_type: DamageElementalType = DamageElementalType.PYRO
    back_damage: int = 2
    cost: Cost = Cost(
        elemental_dice_color = DieColor.PYRO,
        elemental_dice_number = 3,
        charge = 2
    )


# Talents


class BunnyTriggered(SkillTalent):
    name: Literal['Bunny Triggered']
    desc: str = (
        'Combat Action: When your active character is Amber, equip this card. '
        'After Amber equips this card, immediately use Explosive Puppet once. '
        'After you use a Normal Attack: If this card and Baron Bunny are '
        'still on the field, then Baron Bunny explodes and deals 3 Pyro DMG.'
    )
    version: Literal['3.7'] = '3.7'
    charactor_name: Literal['Amber'] = 'Amber'
    cost: Cost = Cost(
        elemental_dice_color = DieColor.PYRO,
        elemental_dice_number = 3
    )
    skill: ExplosivePuppet = ExplosivePuppet()


# charactor base


class Amber(CharactorBase):
    name: Literal['Amber']
    version: Literal['3.7'] = '3.7'
    desc: str = '''"Outrider" Amber'''
    element: ElementType = ElementType.PYRO
    max_hp: int = 10
    max_charge: int = 2
    skills: List[
        Sharpshooter | ExplosivePuppet | FieryRain
    ] = []
    faction: List[FactionType] = [
        FactionType.MONDSTADT
    ]
    weapon_type: WeaponType = WeaponType.BOW
    talent: BunnyTriggered | None = None

    def _init_skills(self) -> None:
        self.skills = [
            Sharpshooter(),
            ExplosivePuppet(),
            FieryRain()
        ]



from .summon import Summons as OtherSummons
from .charactor import (  # noqa: F401
    Charactors, SummonsOfCharactors, CharactorTalents
)
from .status import CharactorStatus, TeamStatus  # noqa: F401
from .card import Cards as OtherCards
from .card.support import Supports  # noqa: F401


# For summons and cards, some will implement in charactor files.
# For status, it is impossible, so no need to collect from other folder.
Summons = OtherSummons | SummonsOfCharactors
Cards = OtherCards | CharactorTalents



from typing import List, Literal, Any, get_origin, get_type_hints

from pydantic import validator

from ..object_base import ObjectBase

from ..consts import (
    ELEMENT_TO_DAMAGE_TYPE, ElementType, ElementalReactionType, IconType, 
    ObjectPositionType, ObjectType, DamageElementalType, DamageType
)
from ..event import (
    MakeDamageEventArguments,
    ReceiveDamageEventArguments,
    RoundEndEventArguments,
    ChangeObjectUsageEventArguments,
)
from ..action import (
    Actions, MakeDamageAction, RemoveObjectAction
)
from ..modifiable_values import DamageDecreaseValue, DamageValue
from ..struct import Cost


class SummonBase(ObjectBase):
    type: Literal[ObjectType.SUMMON] = ObjectType.SUMMON
    renew_type: Literal['ADD', 'RESET', 'RESET_WITH_MAX'] = 'RESET_WITH_MAX'
    name: str
    desc: str
    strict_version_validation: bool = False  # default accept higher versions
    version: str
    usage: int
    max_usage: int

    # icon type is used to show the icon on the summon top right. 
    icon_type: Literal[
        IconType.SHIELD, IconType.BARRIER, IconType.TIMESTATE, IconType.COUNTER
    ]
    # when status icon type is not none, it will show in team status area
    status_icon_type: Literal[IconType.NONE] = IconType.NONE

    @validator('version', pre = True)
    def accept_same_or_higher_version(cls, v: str, values):  # pragma: no cover
        msg = 'version annotation must be Literal with one str'
        if not isinstance(v, str):
            raise NotImplementedError(msg)
        version_hints = get_type_hints(cls)['version']
        if get_origin(version_hints) != Literal:
            raise NotImplementedError(msg)
        version_hints = version_hints.__args__
        if len(version_hints) > 1:
            raise NotImplementedError(msg)
        version_hint = version_hints[0]
        if values['strict_version_validation'] and v != version_hint:
            raise ValueError(
                f'version {v} is not equal to {version_hint}'
            )
        if v < version_hint:
            raise ValueError(
                f'version {v} is lower than {version_hint}'
            )
        return version_hint

    def renew(self, new_status: 'SummonBase') -> None:
        """
        Renew the status. 
        """
        if self.renew_type == 'ADD':
            self.usage += new_status.usage
            if self.max_usage < self.usage:
                self.usage = self.max_usage
        elif self.renew_type == 'RESET':
            raise NotImplementedError('RESET is not supported')
            self.usage = new_status.usage
        else:
            assert self.renew_type == 'RESET_WITH_MAX'
            self.usage = max(self.usage, new_status.usage)

    def is_valid(self, match: Any) -> bool:
        """
        For summons, it is expected to never be used as card.
        """
        raise AssertionError('SummonBase is not expected to be used as card')


class AttackerSummonBase(SummonBase):
    """
    Attacker summons, e.g. Guoba, Oz. They do attack on round end, and 
    disappears when run out of usage. Specially, Melody Loop can also be a
    Attacker Summon, as it also makes damage with type HEAL.
    """
    name: str
    desc: str = '''End Phase: Deal _DAMAGE_ _ELEMENT_ DMG.'''
    version: str
    usage: int
    max_usage: int
    damage_elemental_type: DamageElementalType
    damage: int

    icon_type: Literal[IconType.TIMESTATE] = IconType.TIMESTATE

    def __init__(self, *argv, **kwargs) -> None:
        super().__init__(*argv, **kwargs)
        self.desc = self.desc.replace(
            '_DAMAGE_', str(self.damage)
        ).replace(
            '_ELEMENT_', self.damage_elemental_type.name.capitalize()
        )

    def event_handler_ROUND_END(
        self, event: RoundEndEventArguments, match: Any
    ) -> List[MakeDamageAction]:
        """
        When round end, make damage to the opponent.
        """
        player_idx = self.position.player_idx
        assert self.usage > 0
        self.usage -= 1
        target_table = match.player_tables[1 - player_idx]
        target_charactor = target_table.get_active_charactor()
        return [
            MakeDamageAction(
                source_player_idx = player_idx,
                target_player_idx = 1 - player_idx,
                damage_value_list = [
                    DamageValue(
                        position = self.position,
                        damage_type = DamageType.DAMAGE,
                        target_position = target_charactor.position,
                        damage = self.damage,
                        damage_elemental_type = self.damage_elemental_type,
                        cost = Cost(),
                    )
                ],
            )
        ]

    def _remove(self, match: Any) -> List[RemoveObjectAction]:
        """
        Remove the summon.
        """
        return [
            RemoveObjectAction(
                object_position = self.position,
            )
        ]

    def event_handler_CHANGE_OBJECT_USAGE(
            self, event: ChangeObjectUsageEventArguments, match: Any) \
            -> List[RemoveObjectAction]:
        """
        When usage is 0, remove the summon.
        """
        if self.usage <= 0:
            return self._remove(match)
        return []

    def event_handler_MAKE_DAMAGE(
        self, event: MakeDamageEventArguments, match: Any
    ) -> List[RemoveObjectAction]:
        """
        When usage is 0, remove the summon.
        """
        if self.usage <= 0:
            return self._remove(match)
        return []


class AOESummonBase(AttackerSummonBase):
    """
    Base class that deals AOE damage. It will attack active charactor 
    damage+element, back chractor back_damage_piercing.
    """
    desc: str = (
        'Deals _ACTIVE_ _ELEMENT_ DMG, deals _BACK_ Piercing DMG to all '
        'opposing characters on standby.'
    )
    back_damage: int

    def __init__(self, *argv, **kwargs):
        super().__init__(*argv, **kwargs)
        self.desc = self.desc.replace(
            '_ELEMENT_', self.damage_elemental_type.value.lower().capitalize())
        self.desc = self.desc.replace('_ACTIVE_', str(self.damage))
        self.desc = self.desc.replace('_BACK_', str(self.back_damage))

    def event_handler_ROUND_END(
        self, event: RoundEndEventArguments, match: Any
    ) -> List[MakeDamageAction]:
        ret = super().event_handler_ROUND_END(event, match)
        assert len(ret) == 1
        assert ret[0].type == 'MAKE_DAMAGE'
        target_table = match.player_tables[1 - self.position.player_idx]
        charactors = target_table.charactors
        for cid, charactor in enumerate(charactors):
            if cid == target_table.active_charactor_idx:
                continue
            if charactor.is_defeated:
                continue
            ret[0].damage_value_list.append(
                DamageValue(
                    position = self.position,
                    damage_type = DamageType.DAMAGE,
                    target_position = charactor.position,
                    damage = self.back_damage,
                    damage_elemental_type = DamageElementalType.PIERCING,
                    cost = Cost(),
                )
            )
        return ret


class DefendSummonBase(SummonBase):
    """
    Defend summons, e.g. Ushi, Frog, Baron Bunny. decrease damage with its rule
    when receive damage, and decrease usage by 1, and will do one-time damage 
    in round end.

    Args:
        name: The name of the summon.
        desc: The description of the summon.
        usage: The usage of the summon, which is shown on top right when it is
            summoned, represents the number of times it can decrease damage.
        max_usage: The maximum usage of the summon.
        damage_elemental_type: The elemental type when attack.
        damage: The damage when attack.
        attack_until_run_out_of_usage: Whether attack until run out of usage.
            If true, when usage is not zero, it will remain on the summon area
            and do nothing in round end. If false, when usage is not zero, it
            will do damage in round end and disappear.
        min_damage_to_trigger: The minimum damage to trigger the damage 
            decrease.
        max_in_one_time: the maximum damage decrease in one time.
        decrease_usage_by_damage: Always be false for defend summons.
    """
    name: str
    desc: str
    version: str
    usage: int
    max_usage: int

    # attack params
    damage_elemental_type: DamageElementalType
    damage: int

    # defend params
    attack_until_run_out_of_usage: bool
    min_damage_to_trigger: int
    max_in_one_time: int
    decrease_usage_by_damage: bool = False

    icon_type: Literal[IconType.BARRIER] = IconType.BARRIER

    def event_handler_ROUND_END(
        self, event: RoundEndEventArguments, match: Any
    ) -> List[MakeDamageAction | RemoveObjectAction]:
        """
        When round end, make damage to the opponent if is needed and remove
        itself.
        """
        player_idx = self.position.player_idx
        if self.usage > 0 and self.attack_until_run_out_of_usage:
            # attack until run out of usage
            return []
        target_table = match.player_tables[1 - player_idx]
        target_charactor = target_table.get_active_charactor()
        return [
            MakeDamageAction(
                source_player_idx = player_idx,
                target_player_idx = 1 - player_idx,
                damage_value_list = [
                    DamageValue(
                        position = self.position,
                        damage_type = DamageType.DAMAGE,
                        target_position = target_charactor.position,
                        damage = self.damage,
                        damage_elemental_type = self.damage_elemental_type,
                        cost = Cost(),
                    )
                ],
            )
        ] + self._remove(match)

    def _remove(self, match: Any) -> List[RemoveObjectAction]:
        """
        Remove the summon.
        """
        return [
            RemoveObjectAction(
                object_position = self.position,
            )
        ]

    def value_modifier_DAMAGE_DECREASE(
            self, value: DamageDecreaseValue, match: Any,
            mode: Literal['TEST', 'REAL']) -> DamageDecreaseValue:
        """
        Decrease damage with its rule, and decrease usage.
        """
        if not value.is_corresponding_charactor_receive_damage(
            self.position, match,
        ):
            # not this charactor receive damage, not modify
            return value
        new_usage = value.apply_shield(
            self.usage, self.min_damage_to_trigger, self.max_in_one_time,
            self.decrease_usage_by_damage
        )
        assert mode == 'REAL'
        self.usage = new_usage
        return value


class ShieldSummonBase(DefendSummonBase):
    """
    Shield summons, which decrease damage like yellow shield, decrease damage 
    by its usage and decrease corresponding usage. Currently no such summon.

    Args:
        name: The name of the summon.
        desc: The description of the summon.
        usage: The usage of the summon, which is shown on top right when it is
            summoned, represents the number of times it can decrease damage.
        max_usage: The maximum usage of the summon.
        damage_elemental_type: The elemental type when attack.
        damage: The damage when attack.
        attack_until_run_out_of_usage: Whether attack until run out of usage.
            If true, when usage is not zero, it will remain on the summon area
            and do nothing in round end. If false, when usage is not zero, it
            will do damage in round end and disappear.

        decrease_usage_by_damage: Always be true for shield summons.
        min_damage_to_trigger: Not used, always be 0 for shield summons.
        max_in_one_time: Not used, always be 0 for shield summons.
    """
    name: str
    desc: str
    version: str
    usage: int
    max_usage: int
    damage_elemental_type: DamageElementalType
    damage: int
    attack_until_run_out_of_usage: bool

    icon_type: Literal[IconType.SHIELD] = IconType.SHIELD

    # papams passing to apply_shield is fixed
    min_damage_to_trigger: int = 0
    max_in_one_time: int = 0
    decrease_usage_by_damage: bool = True


class SwirlChangeSummonBase(AttackerSummonBase):
    """
    Base class for summons that can change damage elemental type when swirl
    reaction made by our charactor or summon. 
    Note: create summon before attack, so it can change element immediately.
    """
    name: str
    desc: str = (
        'After your character or Summon triggers a Swirl reaction: Convert '
        'the Elemental Type of this card and change its DMG dealt to the '
        'element Swirled. (Can only be converted once before leaving the '
        'field)'
    )
    version: str
    usage: int
    max_usage: int

    damage_elemental_type: DamageElementalType = DamageElementalType.ANEMO

    def renew(self, new_status: SummonBase) -> None:
        """
        When renew, also renew damage elemental type.
        """
        super().renew(new_status)
        self.damage_elemental_type = DamageElementalType.ANEMO

    def event_handler_RECEIVE_DAMAGE(
        self, event: ReceiveDamageEventArguments, match: Any
    ) -> List[Actions]:
        """
        When anyone receives damage, and has swirl reaction, and made by
        our charactor or summon, and current damage type is anemo, 
        change stormeye damage element.
        """
        if self.damage_elemental_type != DamageElementalType.ANEMO:
            # already changed, do nothing
            return []
        if event.final_damage.position.area not in [
            ObjectPositionType.SKILL, ObjectPositionType.SUMMON
        ]:  # pragma: no cover
            # not charactor or summon made damage, do nothing
            return []
        if event.final_damage.position.player_idx != self.position.player_idx:
            # not our player made damage, do nothing
            return []
        if event.final_damage.element_reaction != ElementalReactionType.SWIRL:
            # not swirl reaction, do nothing
            return []
        elements = event.final_damage.reacted_elements
        assert elements[0] == ElementType.ANEMO, 'First element must be anemo'
        # do type change
        self.damage_elemental_type = ELEMENT_TO_DAMAGE_TYPE[elements[1]]
        return []



from typing import Any, List, Literal

from ...summon.base import AttackerSummonBase

from ...modifiable_values import DamageValue
from ...event import (
    CreateObjectEventArguments, RoundEndEventArguments, 
    RoundPrepareEventArguments
)

from ...action import (
    Actions, CreateObjectAction, MakeDamageAction, RemoveObjectAction
)
from ...struct import Cost, ObjectPosition

from ...consts import (
    DamageElementalType, DamageType, DieColor, 
    ElementType, FactionType, ObjectPositionType, WeaponType
)
from ..charactor_base import (
    ElementalBurstBase, ElementalSkillBase, 
    PhysicalNormalAttackBase, CharactorBase, SkillTalent
)


# Summons


class FierySanctumField(AttackerSummonBase):
    name: Literal['Fiery Sanctum Field'] = 'Fiery Sanctum Field'
    desc: str = (
        'End Phase: Deal 1 Pyro DMG. '
        'When this Summon is on the field and Dehya is on standby on your '
        'side, then when your active character takes damage: Decrease DMG '
        'taken by 1, and if Dehya has at least 7 HP, deal 1 Piercing DMG to '
        'her (once per round).'
    )
    version: Literal['4.1'] = '4.1'
    usage: int = 3
    max_usage: int = 3
    damage_elemental_type: DamageElementalType = DamageElementalType.PYRO
    damage: int = 1
    shield_usage: int = 1
    shield_triggered: bool = False

    def _create_status(self, match: Any) -> List[CreateObjectAction]:
        """
        Create status
        """
        return [CreateObjectAction(
            object_name = self.name,
            object_position = ObjectPosition(
                player_idx = self.position.player_idx,
                area = ObjectPositionType.TEAM_STATUS,
                id = 0,
            ),
            object_arguments = {}
        )]

    def event_handler_CREATE_OBJECT(
        self, event: CreateObjectEventArguments, match: Any
    ) -> List[CreateObjectAction]:
        """
        When created object is self, and not renew, also create teams status
        """
        if event.create_result == 'RENEW':
            # renew, do nothing
            return []
        if (
            event.action.object_name == self.name
            and self.position.check_position_valid(
                event.action.object_position, match, player_idx_same = True,
                area_same = True
            )
        ):
            # name same, and position same, is self created
            return self._create_status(match)
        return []

    def event_handler_ROUND_PREPARE(
        self, event: RoundPrepareEventArguments, match: Any
    ) -> List[CreateObjectAction]:
        # re-generate status
        self.shield_usage = 1
        self.shield_triggered = False
        return self._create_status(match)

    def _remove(self, match: Any) -> List[RemoveObjectAction]:
        """
        If self should remove, and has generated status, remove together.
        """
        status = match.player_tables[
            self.position.player_idx].team_status
        target_status = None
        for s in status:
            if s.name == self.name:
                target_status = s
                break
        ret: List[RemoveObjectAction] = [RemoveObjectAction(
            object_position = self.position,
        )]
        if target_status is not None:
            ret.append(RemoveObjectAction(
                object_position = target_status.position
            ))
        return ret

# Skills


class MoltenInferno(ElementalSkillBase):
    name: Literal['Molten Inferno'] = 'Molten Inferno'
    desc: str = (
        'Summons Fiery Sanctum Field. If Fiery Sanctum Field already exists, '
        'then first deal 1 Pyro DMG.'
    )
    damage: int = 1
    damage_type: DamageElementalType = DamageElementalType.PYRO
    cost: Cost = Cost(
        elemental_dice_color = DieColor.PYRO,
        elemental_dice_number = 3
    )

    def get_actions(self, match: Any) -> List[Actions]:
        """
        Attack and create object
        """
        summons = match.player_tables[self.position.player_idx].summons
        summon_name = 'Fiery Sanctum Field'
        summon_exists = False
        for summon in summons:
            if summon.name == summon_name:
                summon_exists = True
                break
        if summon_exists:
            # make damage and create summon
            return super().get_actions(match) + [
                self.create_summon(summon_name),
            ]
        return [
            self.create_summon(summon_name),
            self.charge_self(1)
        ]


class LeonineBite(ElementalBurstBase):
    name: Literal['Leonine Bite'] = 'Leonine Bite'
    desc: str = (
        'Deals 3 Pyro DMG, then performs "Prepare Skill" for Incineration '
        'Drive'
    )
    damage: int = 3
    damage_type: DamageElementalType = DamageElementalType.PYRO
    cost: Cost = Cost(
        elemental_dice_color = DieColor.PYRO,
        elemental_dice_number = 4,
        charge = 2
    )

    def get_actions(self, match: Any) -> List[Actions]:
        return super().get_actions(match) + [
            self.create_charactor_status('Incineration Drive'),
        ]


class IncinerationDrive(ElementalBurstBase):
    name: Literal['Incineration Drive'] = 'Incineration Drive'
    desc: str = (
        'Deals 3 Pyro DMG.'
    )
    damage: int = 3
    damage_type: DamageElementalType = DamageElementalType.PYRO
    cost: Cost = Cost()

    def is_valid(self, match: Any) -> bool:
        """
        Always invalid for prepare skills
        """
        return False


# Talents


class StalwartAndTrue(SkillTalent):
    name: Literal['Stalwart and True']
    desc: str = (
        'Combat Action: When your active character is Dehya, equip this card. '
        'After Dehya equips this card, immediately use Molten Inferno once. '
        'End Phase: If your Dehya, who has this card equipped, has no more '
        'than 6 HP, heal that character for 2 HP.'
    )
    version: Literal['4.1'] = '4.1'
    charactor_name: Literal['Dehya'] = 'Dehya'
    cost: Cost = Cost(
        elemental_dice_color = DieColor.PYRO,
        elemental_dice_number = 4,
    )
    skill: MoltenInferno = MoltenInferno()

    def event_handler_ROUND_END(
        self, event: RoundEndEventArguments, match: Any
    ) -> List[Actions]:
        """
        If equipped, and dehya hp <= 6, heal 2 hp
        """
        if self.position.area != ObjectPositionType.CHARACTOR:
            # not equipped
            return []
        charactor = match.player_tables[
            self.position.player_idx].charactors[self.position.charactor_idx]
        if charactor.hp <= 6:
            return [MakeDamageAction(
                source_player_idx = self.position.player_idx,
                target_player_idx = self.position.player_idx,
                damage_value_list = [
                    DamageValue(
                        position = self.position,
                        damage_type = DamageType.HEAL,
                        target_position = charactor.position,
                        damage = -2,
                        damage_elemental_type = DamageElementalType.HEAL,
                        cost = self.cost.copy()
                    )
                ],
            )]
        return []


# charactor base


class Dehya(CharactorBase):
    name: Literal['Dehya']
    version: Literal['4.1'] = '4.1'
    desc: str = '''"Flame-Mane" Dehya'''
    element: ElementType = ElementType.PYRO
    max_hp: int = 10
    max_charge: int = 2
    skills: List[
        PhysicalNormalAttackBase | MoltenInferno | LeonineBite 
        | IncinerationDrive
    ] = []
    faction: List[FactionType] = [
        FactionType.SUMERU
    ]
    weapon_type: WeaponType = WeaponType.CLAYMORE
    talent: StalwartAndTrue | None = None

    def _init_skills(self) -> None:
        self.skills = [
            PhysicalNormalAttackBase(
                name = 'Sandstorm Assault',
                cost = PhysicalNormalAttackBase.get_cost(self.element),
            ),
            MoltenInferno(),
            LeonineBite(),
            IncinerationDrive(),
        ]



import pytest
from src.lpsim.server.deck import Deck


def test_create_mob():
    TEST_DECK = {
        'name': 'Deck',
        'charactors': [
            {
                'name': 'PhysicalMob',
                'element': 'ELECTRO',
            },
            {
                'name': 'ElectroMobMage',
                'element': 'ELECTRO',
            },
            {
                'name': 'ElectroMob',
                'element': 'ELECTRO',
            },
            {
                'name': 'DendroMob',
            },
            {
                'name': 'DendroMobMage',
            },
        ],
        'cards': [
            {
                'name': 'Strategize',
            }
        ] * 30,
    }
    Deck(**TEST_DECK)
    TEST_DECK['charactors'][1]['element'] = 'DENDRO'
    with pytest.raises(ValueError):
        Deck(**TEST_DECK)
    TEST_DECK['charactors'][1]['element'] = 'ELECTRO'
    TEST_DECK['charactors'][2]['element'] = 'DENDRO'
    with pytest.raises(ValueError):
        Deck(**TEST_DECK)
    TEST_DECK['charactors'][2]['element'] = 'ELECTRO'



from typing import Any, Literal, List

from ...struct import Cost

from ...modifiable_values import (
    DamageDecreaseValue, DamageElementEnhanceValue, DamageValue
)
from ...consts import (
    ELEMENT_TO_ENCHANT_ICON, DamageElementalType, DamageType, ElementType, 
    IconType, ObjectType, PlayerActionLabels
)
from ..base import StatusBase
from ...action import (
    ActionEndAction, MakeDamageAction, RemoveObjectAction, Actions, 
    SkipPlayerActionAction, UseSkillAction
)
from ...event import (
    ChooseCharactorEventArguments, MakeDamageEventArguments, 
    PlayerActionStartEventArguments, RoundPrepareEventArguments, 
    SwitchCharactorEventArguments
)


class CharactorStatusBase(StatusBase):
    """
    Base class of charactor status.
    """
    type: Literal[ObjectType.CHARACTOR_STATUS] = ObjectType.CHARACTOR_STATUS
    name: str
    desc: str
    version: str
    usage: int
    max_usage: int


class UsageCharactorStatus(CharactorStatusBase):
    """
    Base class of team status that based on usages.
    It will implement a method to check if run out of usage; when run out
    of usage, it will remove itself.

    By default, it listens to MAKE_DAMAGE event to check if run out of usage,
    as most usage-based status will decrease usage when damage made.
    Not using AFTER_MAKE_DAMAGE because we should remove the status before
    side effects of elemental reaction, which is triggered on RECEVIE_DAMAGE.

    If it is not a damage related status, check_should_remove should be
    called manually.
    """
    name: str
    desc: str
    version: str
    usage: int
    max_usage: int

    def check_should_remove(self) -> List[RemoveObjectAction]:
        """
        Check if the status should be removed.
        when usage has changed, call this function to check if the status
        should be removed.
        """
        if self.usage <= 0:
            return [RemoveObjectAction(
                object_position = self.position,
            )]
        return []

    def event_handler_MAKE_DAMAGE(
        self, event: MakeDamageEventArguments, match: Any
    ) -> List[RemoveObjectAction]:
        """
        When damage made, check whether the team status should be removed.
        """
        return self.check_should_remove()


class RoundCharactorStatus(CharactorStatusBase):
    """
    Base class of team status that based on rounds.  It will implement the 
    event trigger on ROUND_PREPARE to check if run out of usages; when run out
    of usages, it will remove itself.
    """
    name: str
    desc: str
    version: str
    usage: int
    max_usage: int

    def check_should_remove(self) -> List[RemoveObjectAction]:
        """
        Check if the status should be removed.
        when round has changed, call this function to check if the status
        should be removed.
        """
        if self.usage <= 0:
            return [RemoveObjectAction(
                object_position = self.position,
            )]
        return []

    def event_handler_ROUND_PREPARE(
        self, event: RoundPrepareEventArguments, match: Any
    ) -> List[Actions]:
        """
        When reaching prepare, decrease usage and check whether the team 
        status should be removed.
        """
        self.usage -= 1
        return list(self.check_should_remove())


class DefendCharactorStatus(UsageCharactorStatus):
    """
    Base class of defend status (purple shield), decrease damage with its rule
    when receive damage, and decrease usage by 1.
    """
    name: str
    desc: str
    version: str
    usage: int
    max_usage: int
    min_damage_to_trigger: int
    max_in_one_time: int
    decrease_usage_by_damage: bool = False

    icon_type: Literal[IconType.BARRIER] = IconType.BARRIER

    def value_modifier_DAMAGE_DECREASE(
            self, value: DamageDecreaseValue, match: Any,
            mode: Literal['TEST', 'REAL']) -> DamageDecreaseValue:
        """
        Decrease damage with its rule, and decrease 1 usage.
        """
        if not value.is_corresponding_charactor_receive_damage(
            self.position, match,
        ):
            # not this charactor receive damage, not modify
            return value
        if self.usage <= 0:
            return value
        new_usage = value.apply_shield(
            self.usage, self.min_damage_to_trigger, 
            self.max_in_one_time, self.decrease_usage_by_damage,
        )
        assert mode == 'REAL'
        self.usage = new_usage
        return value


class ShieldCharactorStatus(DefendCharactorStatus):
    """
    Base class of shield status (yellow shield), decrease damage by its usage
    and decrease corresponding usage.
    """
    name: str
    desc: str
    version: str
    usage: int
    max_usage: int

    icon_type: Literal[IconType.SHIELD] = IconType.SHIELD

    # papams passing to apply_shield is fixed
    min_damage_to_trigger: int = 0
    max_in_one_time: int = 0
    decrease_usage_by_damage: bool = True


class ElementalInfusionCharactorStatus(CharactorStatusBase):
    """
    Base class of doing elemental infusion. It will infuse all physical damage
    to corresponding elemental damage for this charactor. If there extra
    conditions, filter the condition and call value modifier in subclasses.
    """

    name: Literal[
        'Pyro Elemental Infusion',
        'Hydro Elemental Infusion',
        'Electro Elemental Infusion',
        'Cryo Elemental Infusion',
        'Anemo Elemental Infusion',
        'Geo Elemental Infusion',
        'Dendro Elemental Infusion',
    ]
    desc: str = (
        'When the charactor to which it is attached to deals Physical Damage, '
        'it will be turned into XXX DMG. '
    )
    infused_elemental_type: DamageElementalType = DamageElementalType.PHYSICAL

    icon_type: Literal[
        IconType.ELEMENT_ENCHANT_ELEC,
        IconType.ELEMENT_ENCHANT_FIRE,
        IconType.ELEMENT_ENCHANT_WATER,
        IconType.ELEMENT_ENCHANT_ICE,
        IconType.ELEMENT_ENCHANT_WIND,
        IconType.ELEMENT_ENCHANT_ROCK,
        IconType.ELEMENT_ENCHANT_GRASS,
        IconType.ATK_UP,
    ] = IconType.ATK_UP

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if self.infused_elemental_type == DamageElementalType.PHYSICAL:
            # not set elemental type manually, get it from name
            element = self.name.split(' ')[0].upper()
            assert element in [
                'PYRO', 'HYDRO', 'ELECTRO', 'CRYO', 'ANEMO', 'GEO', 'DENDRO'
            ], ('In ElementalInfusion: element not set right value')
            assert self.name.split(' ')[1:] == ['Elemental', 'Infusion']
            self.infused_elemental_type = DamageElementalType(element)
            self.desc = self.desc.replace('XXX', element)
        if self.icon_type == IconType.ATK_UP:
            self.icon_type = ELEMENT_TO_ENCHANT_ICON[
                ElementType(self.infused_elemental_type.value)]  # type: ignore

    def value_modifier_DAMAGE_ELEMENT_ENHANCE(
        self, value: DamageElementEnhanceValue, match: Any,
        mode: Literal['TEST', 'REAL'],
    ) -> DamageElementEnhanceValue:
        """
        When self use skill, change physical to corresponding element.
        NOTE: this function will not change the usage.
        """
        assert mode == 'REAL'
        if not value.is_corresponding_charactor_use_damage_skill(
            self.position, match, None
        ):
            # not corresponding charactor use skill, do nothing
            return value
        if self.usage <= 0:  # pragma: no cover
            # no usage, do nothing
            return value
        # change physical to other element
        if value.damage_elemental_type == DamageElementalType.PHYSICAL:
            value.damage_elemental_type = self.infused_elemental_type
        return value


class PrepareCharactorStatus(CharactorStatusBase):
    """
    Prepare status will trigger when self player action start, and try to
    use a specified skill, and skip next player action phase.
    if this charactor is stunned, it will do nothing.

    When this charactor is not active charactor, it will remove itself. 
    """
    name: str
    desc: str
    version: str
    charactor_name: str
    skill_name: str

    usage: int = 1
    max_usage: int = 1
    icon_type: Literal[IconType.SPECIAL] = IconType.SPECIAL

    def event_handler_SWITCH_CHARACTOR(
        self, event: SwitchCharactorEventArguments, match: Any
    ) -> List[RemoveObjectAction]:
        """
        If self switch charactor, after switch if this charactor is not active
        charactor, remove this status.
        """
        if event.action.player_idx != self.position.player_idx:
            # not self switch charactor
            return []
        return [RemoveObjectAction(
            object_position = self.position,
        )]

    def event_handler_CHOOSE_CHARACTOR(
        self, event: ChooseCharactorEventArguments, match: Any
    ) -> List[MakeDamageAction]:
        """
        As choose charactor only happens on charactor defeated, it is not
        possible.
        """
        if event.action.player_idx != self.position.player_idx:
            # not self choose charactor
            return []
        raise NotImplementedError('Should not be called')
        return [RemoveObjectAction(
            object_position = self.position,
        )]

    def event_handler_PLAYER_ACTION_START(
        self, event: PlayerActionStartEventArguments, match: Any
    ) -> List[RemoveObjectAction | UseSkillAction 
              | SkipPlayerActionAction | ActionEndAction]:
        """
        If self player action start, and this charactor is at front,
        use skill, remove this status, and skip player action.
        """
        if event.player_idx != self.position.player_idx:
            # not self player action start, do nothing
            return []
        table = match.player_tables[self.position.player_idx]
        active_idx = table.active_charactor_idx
        if active_idx != self.position.charactor_idx:  # pragma: no cover
            # not active charactor, do nothing
            return []
        charactor = table.get_active_charactor()
        if charactor.is_stunned:
            # stunned, do nothing
            return []
        ret: List[
            RemoveObjectAction | UseSkillAction | SkipPlayerActionAction
            | ActionEndAction
        ] = []
        ret.append(RemoveObjectAction(object_position = self.position))
        assert charactor.name == self.charactor_name
        # use skill
        for skill in charactor.skills:
            if skill.name == self.skill_name:
                # clear, should not remove self first
                ret.clear()
                ret.append(UseSkillAction(skill_position = skill.position))
                # skip player action
                ret.append(SkipPlayerActionAction())
                # remove self
                ret.append(RemoveObjectAction(object_position = self.position))
                # action end action to switch player
                ret.append(ActionEndAction(
                    action_label = PlayerActionLabels.SKILL,
                    do_combat_action = True,
                    position = skill.position,
                ))
                # note based on its description, no skill end action will be
                # triggered.
                return ret
        else:
            raise AssertionError('Skill not found')


class ReviveCharactorStatus(UsageCharactorStatus):
    name: str
    version: str
    heal: int

    desc: str = (
        'When the character to which this is attached would be defeated: '
        'Remove this effect, ensure the character will not be defeated, and '
        'heal them to _HEAL_ HP.'
    )
    usage: int = 1
    max_usage: int = 1
    icon_type: Literal[IconType.REVIVE] = IconType.REVIVE

    def event_handler_MAKE_DAMAGE(
        self, event: MakeDamageEventArguments, match: Any
    ) -> List[MakeDamageAction | RemoveObjectAction]:
        """
        when make damage, check whether this charactor's hp is 0. if so,
        heal it by self.heal hp.

        Using make damage action, so it will trigger immediately before
        receiving any other damage, and can be defeated by other damage.
        """
        charactor = match.player_tables[self.position.player_idx].charactors[
            self.position.charactor_idx]
        if charactor.hp > 0:
            # hp not 0, do nothing
            return []
        if self.usage <= 0:  # pragma: no cover
            # no usage, do nothing
            return []
        # heal this charactor by 1
        self.usage -= 1
        return [MakeDamageAction(
            source_player_idx = self.position.player_idx,
            target_player_idx = self.position.player_idx,
            damage_value_list = [
                DamageValue(
                    position = self.position,
                    damage_type = DamageType.HEAL,
                    target_position = charactor.position,
                    damage = - self.heal,
                    damage_elemental_type = DamageElementalType.HEAL,
                    cost = Cost()
                )
            ],
        )] + self.check_should_remove()



from typing import Any, List, Literal

from ...summon.base import AttackerSummonBase

from ...modifiable_values import DamageValue
from ...event import SkillEndEventArguments

from ...action import (
    Actions, MakeDamageAction, SwitchCharactorAction
)
from ...struct import Cost

from ...consts import (
    DamageElementalType, DamageType, DieColor, ElementType, FactionType, 
    ObjectPositionType, SkillType, WeaponType
)
from ..charactor_base import (
    ElementalBurstBase, ElementalSkillBase, 
    PhysicalNormalAttackBase, CharactorBase, SkillBase, SkillTalent
)


# Summons


class ShadowswordLoneGale(AttackerSummonBase):
    name: Literal['Shadowsword: Lone Gale']
    desc: str = '''End Phase: Deal 1 Anemo DMG. Usage(s): 2'''
    version: Literal['3.3'] = '3.3'
    usage: int = 2
    max_usage: int = 2
    damage_elemental_type: DamageElementalType = DamageElementalType.ANEMO
    damage: int = 1

    def event_handler_SKILL_END(
        self, event: SkillEndEventArguments, match: Any
    ) -> List[MakeDamageAction]:
        """
        If kenki use elemental burst, make damage
        """
        if not self.position.check_position_valid(
            event.action.position, match, player_idx_same = True,
            target_area = ObjectPositionType.SKILL
        ):
            # not our player skill made damage, do nothing
            return []
        if event.action.skill_type != SkillType.ELEMENTAL_BURST:
            # not elemental burst, do nothing
            return []
        charactor = match.player_tables[self.position.player_idx].charactors[
            event.action.position.charactor_idx]
        if charactor.name != 'Maguu Kenki':
            # not kenki, do nothing
            return []
        target_charactor = match.player_tables[
            1 - self.position.player_idx].get_active_charactor()
        return [MakeDamageAction(
            source_player_idx = self.position.player_idx,
            target_player_idx = 1 - self.position.player_idx,
            damage_value_list = [
                DamageValue(
                    position = self.position,
                    damage_type = DamageType.DAMAGE,
                    target_position = target_charactor.position,
                    damage = self.damage,
                    damage_elemental_type = self.damage_elemental_type,
                    cost = Cost(),
                )
            ]
        )]


class ShadowswordGallopingFrost(ShadowswordLoneGale):
    name: Literal['Shadowsword: Galloping Frost']
    desc: str = '''End Phase: Deal 1 Cryo DMG. Usage(s): 2'''
    version: Literal['3.3'] = '3.3'
    usage: int = 2
    max_usage: int = 2
    damage_elemental_type: DamageElementalType = DamageElementalType.CRYO
    damage: int = 1


# Skills


class BlusteringBlade(ElementalSkillBase):
    name: Literal['Blustering Blade'] = 'Blustering Blade'
    desc: str = '''Summons 1 Shadowsword: Lone Gale.'''
    damage: int = 0
    damage_type: DamageElementalType = DamageElementalType.ANEMO
    cost: Cost = Cost(
        elemental_dice_color = DieColor.ANEMO,
        elemental_dice_number = 3
    )

    def get_actions(self, match: Any) -> List[Actions]:
        """
        only create object
        """
        return [
            self.create_summon('Shadowsword: Lone Gale'),
            self.charge_self(1)
        ]


class FrostyAssault(ElementalSkillBase):
    name: Literal['Frosty Assault'] = 'Frosty Assault'
    desc: str = '''Summons 1 Shadowsword: Galloping Frost.'''
    damage: int = 0
    damage_type: DamageElementalType = DamageElementalType.CRYO
    cost: Cost = Cost(
        elemental_dice_color = DieColor.CRYO,
        elemental_dice_number = 3
    )

    def get_actions(self, match: Any) -> List[Actions]:
        """
        only create object
        """
        return [
            self.create_summon('Shadowsword: Galloping Frost'),
            self.charge_self(1)
        ]


class PseudoTenguSweeper(ElementalBurstBase):
    name: Literal['Pseudo Tengu Sweeper'] = 'Pseudo Tengu Sweeper'
    desc: str = (
        'Deals 4 Anemo DMG, triggers the effect(s) of all your Shadowsword '
        'Summon(s). (Does not consume their Usages)'
    )
    damage: int = 4
    damage_type: DamageElementalType = DamageElementalType.ANEMO
    cost: Cost = Cost(
        elemental_dice_color = DieColor.ANEMO,
        elemental_dice_number = 3,
        charge = 3
    )


# Talents


class TranscendentAutomaton(SkillTalent):
    name: Literal['Transcendent Automaton']
    desc: str = (
        'Combat Action: When your active character is Maguu Kenki, equip this '
        'card. After Maguu Kenki equips this card, immediately use Blustering '
        'Blade once. After your Maguu Kenki, who has this card equipped, uses '
        'Blustering Blade, you will switch to your next character. You will '
        'switch to your previous character when your Maguu Kenki, who has '
        'this card equipped, uses Frosty Assault.'
    )
    version: Literal['3.3'] = '3.3'
    charactor_name: Literal['Maguu Kenki'] = 'Maguu Kenki'
    cost: Cost = Cost(
        elemental_dice_color = DieColor.ANEMO,
        elemental_dice_number = 3,
    )
    skill: BlusteringBlade = BlusteringBlade()

    def event_handler_SKILL_END(
        self, event: SkillEndEventArguments, match: Any
    ) -> List[SwitchCharactorAction]:
        """
        When self use elemental skill, swich to previous or next charactor
        """
        if not self.position.check_position_valid(
            event.action.position, match, player_idx_same = True,
            charactor_idx_same = True, 
            source_area = ObjectPositionType.CHARACTOR,
            target_area = ObjectPositionType.SKILL,
        ):
            # not equipped on charactor, or skill source not self
            return []
        if event.action.skill_type != SkillType.ELEMENTAL_SKILL:
            # not elemental skill, do nothing
            return []
        table = match.player_tables[self.position.player_idx]
        skill: SkillBase = match.get_object(event.action.position)
        if skill.name == 'Blustering Blade':
            # switch to next charactor
            next_idx = table.next_charactor_idx()
            if next_idx is None:
                # no next charactor, do nothing
                return []
            return [SwitchCharactorAction(
                player_idx = self.position.player_idx,
                charactor_idx = next_idx,
            )]
        elif skill.name == 'Frosty Assault':
            # switch to previous charactor
            prev_idx = table.previous_charactor_idx()
            if prev_idx is None:
                # no next charactor, do nothing
                return []
            return [SwitchCharactorAction(
                player_idx = self.position.player_idx,
                charactor_idx = prev_idx,
            )]
        else:
            raise AssertionError('skill name not match')


# charactor base


class MaguuKenki(CharactorBase):
    name: Literal['Maguu Kenki']
    version: Literal['3.4'] = '3.4'
    desc: str = '''"Ingenious Machine" Maguu Kenki'''
    element: ElementType = ElementType.ANEMO
    max_hp: int = 10
    max_charge: int = 3
    skills: List[
        PhysicalNormalAttackBase
        | BlusteringBlade | FrostyAssault | PseudoTenguSweeper
    ] = []
    faction: List[FactionType] = [
        FactionType.MONSTER
    ]
    weapon_type: WeaponType = WeaponType.OTHER
    talent: TranscendentAutomaton | None = None

    def _init_skills(self) -> None:
        self.skills = [
            PhysicalNormalAttackBase(
                name = 'Ichimonji',
                cost = PhysicalNormalAttackBase.get_cost(self.element),
            ),
            BlusteringBlade(),
            FrostyAssault(),
            PseudoTenguSweeper()
        ]



"""
Event handlers to implement system controls and special rules 
"""
from typing import List, Literal, Any
from .object_base import ObjectBase
from .struct import ObjectPosition
from .consts import ObjectPositionType, DieColor
from .event import (
    ReceiveDamageEventArguments,
    AfterMakeDamageEventArguments,
    CharactorDefeatedEventArguments,
    RoundEndEventArguments,
    DrawCardEventArguments,
)
from .action import (
    Actions,
    CharactorDefeatedAction,
    GenerateChooseCharactorRequestAction,
    DrawCardAction,
    RemoveCardAction
)
from .modifiable_values import InitialDiceColorValue, RerollValue
from .elemental_reaction import elemental_reaction_side_effect


class SystemEventHandlerBase(ObjectBase):
    """
    Base class of system event handlers. Its position should be SYSTEM.
    """

    name: Literal['SystemEventHandlerBase']

    position: ObjectPosition = ObjectPosition(
        player_idx = -1,
        area = ObjectPositionType.SYSTEM,
        id = -1,
    )

    def renew(
        self, target: 'SystemEventHandlerBase'
    ) -> None:  # pragma: no cover
        """
        No need to renew.
        """
        pass


class SystemEventHandler(SystemEventHandlerBase):

    name: Literal['System'] = 'System'
    version: Literal['3.3', '3.4'] = '3.4'

    def event_handler_DRAW_CARD(
            self, event: DrawCardEventArguments, match: Any) -> List[Actions]:
        """
        After draw card, check whether max card number is exceeded.
        """
        player_idx = event.action.player_idx
        hand_size = event.hand_size
        max_hand_size = event.max_hand_size
        actions = []
        hands = match.player_tables[player_idx].hands
        if hand_size > max_hand_size:
            for i in range(hand_size, max_hand_size, -1):
                # remove the last card in hand repeatedly until hand size
                # is less than max hand size
                actions.append(RemoveCardAction(
                    position = hands[i - 1].position,
                    remove_type = 'BURNED'
                ))
        return actions

    def event_handler_RECEIVE_DAMAGE(
        self, event: ReceiveDamageEventArguments, match: Any
    ) -> List[Actions]:
        """
        After receive damage, generate side effects of elemental reaction.
        """
        reaction = event.elemental_reaction
        player_idx = event.final_damage.target_position.player_idx
        charactor_idx = event.final_damage.target_position.charactor_idx
        act = elemental_reaction_side_effect(
            reaction, player_idx, charactor_idx, 
            version = self.version
        )
        if act is not None:
            return [act]
        return []

    def event_handler_AFTER_MAKE_DAMAGE(
        self, event: AfterMakeDamageEventArguments, match: Any
    ) -> List[Actions]:
        """
        After damage made, check whether anyone has defeated.
        """
        actions: List[Actions] = []
        for pnum, table in enumerate(match.player_tables):
            for cnum, charactor in enumerate(table.charactors):
                if charactor.hp == 0 and charactor.is_alive:
                    charactor.is_alive = False  # mark as defeated first
                    actions.append(CharactorDefeatedAction(
                        player_idx = pnum,
                        charactor_idx = cnum,
                    ))
        return actions

    def event_handler_CHARACTOR_DEFEATED(
        self, event: CharactorDefeatedEventArguments, match: Any
    ) -> List[Actions]:
        """
        After charactor defeated, check whether need to switch charactor.
        """
        if event.need_switch:
            return [GenerateChooseCharactorRequestAction(
                player_idx = event.action.player_idx,
            )]
        return []

    def event_handler_ROUND_END(
            self, event: RoundEndEventArguments, match: Any) -> List[Actions]:
        """
        After round end, draw card for each player.
        """
        actions: List[Actions] = []
        first = event.player_go_first
        initial_card_draw = event.initial_card_draw
        actions.append(DrawCardAction(
            player_idx = first,
            number = initial_card_draw,
            draw_if_filtered_not_enough = True
        ))
        actions.append(DrawCardAction(
            player_idx = 1 - first,
            number = initial_card_draw,
            draw_if_filtered_not_enough = True
        ))
        return actions


class OmnipotentGuideEventHandler(SystemEventHandlerBase):

    name: Literal['Omnipotent Guide'] = 'Omnipotent Guide'

    def value_modifier_INITIAL_DICE_COLOR(
            self, value: InitialDiceColorValue, 
            match: Any,
            mode: Literal['REAL', 'TEST']) -> InitialDiceColorValue:
        """
        remove current dice color and add 100 omni dice color
        """
        value.dice_colors = [DieColor.OMNI] * 100
        return value

    def value_modifier_REROLL(
            self, value: RerollValue, 
            match: Any,
            mode: Literal['REAL', 'TEST']) -> RerollValue:
        """
        reroll set to 0
        """
        value.value = 0
        return value


SystemEventHandlers = (
    SystemEventHandlerBase | SystemEventHandler | OmnipotentGuideEventHandler
)



from typing import Any, List, Literal

from ...summon import Summons
from ...summon.base import AttackerSummonBase

from ...event import DeclareRoundEndEventArguments

from ...action import Actions, MakeDamageAction, RemoveObjectAction
from ...struct import Cost

from ...consts import (
    DamageElementalType, DieColor, ElementType, FactionType, WeaponType
)
from ..charactor_base import (
    ElementalBurstBase, ElementalNormalAttackBase, ElementalSkillBase, 
    CharactorBase, SkillTalent
)


# Summons


class SesshouSakura(AttackerSummonBase):
    name: Literal['Sesshou Sakura'] = 'Sesshou Sakura'
    desc: str = (
        'End Phase: Deal 1 Electro DMG. '
        '(Can stack. Max 6 stacks.) '
        'When you declare the end of your Round: If this summon has at least '
        '4 Usage(s) remaining, deal 1 Electro DMG. (Consumes Usage(s))'
    )
    version: Literal['3.7'] = '3.7'
    usage: int = 3
    max_usage: int = 6
    damage_elemental_type: DamageElementalType = DamageElementalType.ELECTRO
    damage: int = 1
    renew_type: Literal['ADD'] = 'ADD'

    def event_handler_DECLARE_ROUND_END(
        self, event: DeclareRoundEndEventArguments, match: Any
    ) -> List[MakeDamageAction]:
        """
        if usage larger than 4, return event_handler_ROUND_END
        """
        if event.action.player_idx != self.position.player_idx:
            # not our player, do nothing
            return []
        if self.usage < 4:
            # not enough usage, do nothing
            return []
        # as attack not use event, use fake event instead.
        fake_event: Any = None
        # make damage
        return self.event_handler_ROUND_END(fake_event, match)


# Skills


class YakanEvocationSesshouSakura(ElementalSkillBase):
    name: Literal[
        'Yakan Evocation: Sesshou Sakura'] = 'Yakan Evocation: Sesshou Sakura'
    desc: str = '''Summons 1 Sesshou Sakura'''
    damage: int = 0
    damage_type: DamageElementalType = DamageElementalType.PIERCING
    cost: Cost = Cost(
        elemental_dice_color = DieColor.ELECTRO,
        elemental_dice_number = 3
    )

    def get_actions(self, match: Any) -> List[Actions]:
        """
        create object
        """
        return [
            self.charge_self(1),
            self.create_summon('Sesshou Sakura'),
        ]


class GreatSecretArtTenkoKenshin(ElementalBurstBase):
    name: Literal[
        'Great Secret Art: Tenko Kenshin'] = 'Great Secret Art: Tenko Kenshin'
    desc: str = (
        'Deals 4 Electro DMG. If your side of the field has a Sesshou Sakura, '
        'destroy it and create Tenko Thunderbolts.'
    )
    damage: int = 4
    damage_type: DamageElementalType = DamageElementalType.ELECTRO
    cost: Cost = Cost(
        elemental_dice_color = DieColor.ELECTRO,
        elemental_dice_number = 3,
        charge = 2
    )

    def get_actions(self, match: Any) -> List[Actions]:
        ret = super().get_actions(match)
        summons = match.player_tables[self.position.player_idx].summons
        sakura: Summons | None = None
        for summon in summons:
            if summon.name == 'Sesshou Sakura':
                sakura = summon
                break
        if sakura is not None:
            # destroy and add status
            ret.append(RemoveObjectAction(object_position = sakura.position))
            ret.append(self.create_team_status('Tenko Thunderbolts'))
            if self.is_talent_equipped(match):
                # add status
                ret.append(self.create_charactor_status(
                    "The Shrine's Sacred Shade"))
        return ret


# Talents


class TheShrinesSacredShade(SkillTalent):
    name: Literal["The Shrine's Sacred Shade"] = "The Shrine's Sacred Shade"
    desc: str = (
        'Combat Action: When your active character is Yae Miko, equip this '
        'card. After Yae Miko equips this card, immediately use Great Secret '
        'Art: Tenko Kenshin once. After your Yae Miko, who has this card '
        'equipped, destroys a Sesshou Sakura through Great Secret Art: Tenko '
        'Kenshin, the next use of Yakan Evocation: Sesshou Sakura this Round '
        'will cost 2 less Elemental Dice.'
    )
    version: Literal['3.7'] = '3.7'
    charactor_name: Literal['Yae Miko'] = 'Yae Miko'
    cost: Cost = Cost(
        elemental_dice_color = DieColor.ELECTRO,
        elemental_dice_number = 3,
        charge = 2
    )
    skill: GreatSecretArtTenkoKenshin = GreatSecretArtTenkoKenshin()


# charactor base


class YaeMiko(CharactorBase):
    name: Literal['Yae Miko']
    version: Literal['3.7'] = '3.7'
    desc: str = '''"Astute Amusement" Yae Miko'''
    element: ElementType = ElementType.ELECTRO
    max_hp: int = 10
    max_charge: int = 2
    skills: List[
        ElementalNormalAttackBase | YakanEvocationSesshouSakura
        | GreatSecretArtTenkoKenshin
    ] = []
    faction: List[FactionType] = [
        FactionType.INAZUMA
    ]
    weapon_type: WeaponType = WeaponType.CATALYST
    talent: TheShrinesSacredShade | None = None

    def _init_skills(self) -> None:
        self.skills = [
            ElementalNormalAttackBase(
                name = 'Spiritfox Sin-Eater',
                damage_type = self.element,
                cost = ElementalNormalAttackBase.get_cost(self.element),
            ),
            YakanEvocationSesshouSakura(),
            GreatSecretArtTenkoKenshin(),
        ]



from typing import Literal

from ...struct import Cost
from .locations import JadeChamber as JC_4_0
from .companions import Katheryne as K_3_6
from .companions import ChefMao as CM_4_1
from .companions import Dunyarzad as D_4_1
from .items import NRE as NRE_4_1


class JadeChamber_3_3(JC_4_0):
    version: Literal['3.3']
    cost: Cost = Cost(same_dice_number = 1)


class Katheryne_3_3(K_3_6):
    version: Literal['3.3']
    cost: Cost = Cost(any_dice_number = 2)


class ChefMao_3_3(CM_4_1):
    desc: str = (
        'After playing a Food Event Card: Create 1 random Elemental Die. '
        '(Once per Round) '
    )
    version: Literal['3.3']
    limited_usage: int = 0


class Dunyarzad_3_7(D_4_1):
    desc: str = (
        'When playing a Companion Support Card: Spend 1 less Elemental Dice. '
        '(Once per Round) '
    )
    version: Literal['3.7']
    limited_usage: int = 0


class NRE_3_3(NRE_4_1):
    version: Literal['3.3']
    cost: Cost = Cost(any_dice_number = 2)


OldVersionLocations = JadeChamber_3_3 | JadeChamber_3_3
OldVersionCompanions = (
    Dunyarzad_3_7

    | Katheryne_3_3 | ChefMao_3_3
)
OldVersionItems = NRE_3_3 | NRE_3_3



from typing import List, Literal
from ..cryo.fatui_cryo_cicin_mage import FatuiCryoCicinMage as FCC_4_1
from ..cryo.fatui_cryo_cicin_mage import CryoCicins as CC_4_1
from ..cryo.fatui_cryo_cicin_mage import MistySummons as MS_4_1
from ..cryo.fatui_cryo_cicin_mage import CicinIcicle, BlizzardBranchBlossom


class CryoCicins_3_7(CC_4_1):
    desc: str = (
        'End Phase: Deal 1 Cryo DMG. (Can stack. Max 3 stacks.) '
        'After Fatui Cryo Cicin Mage performs a Normal Attack: This card '
        'gains 1 Usage(s). After your character takes Elemental Reaction DMG: '
        'This card loses 1 Usage(s).'
    )
    version: Literal['3.7'] = '3.7'
    decrease_only_self_damage: bool = False


class MistySummons(MS_4_1):
    version: Literal['3.7'] = '3.7'


class FatuiCryoCicinMage_3_7(FCC_4_1):
    version: Literal['3.7']
    skills: List[
        CicinIcicle | MistySummons | BlizzardBranchBlossom
    ] = []

    def _init_skills(self) -> None:
        self.skills = [
            CicinIcicle(),
            MistySummons(),
            BlizzardBranchBlossom(),
        ]



from typing import List, Literal

from ..charactor_base import PhysicalNormalAttackBase
from ...struct import Cost
from ...consts import DieColor
from ..electro.beidou import Beidou as Beidou_3_8
from ..electro.beidou import Wavestrider as Wavestrider_3_8
from ..electro.beidou import Stormbreaker as Stormbreaker_3_8
from ..electro.beidou import Tidecaller


class Wavestrider(Wavestrider_3_8):
    damage: int = 2


class Stormbreaker(Stormbreaker_3_8):
    damage: int = 3
    cost: Cost = Cost(
        elemental_dice_color = DieColor.ELECTRO,
        elemental_dice_number = 4,
        charge = 3
    )


class Beidou_3_4(Beidou_3_8):
    version: Literal['3.4']
    skills: List[
        PhysicalNormalAttackBase | Tidecaller | Wavestrider | Stormbreaker
    ] = []

    def _init_skills(self) -> None:
        self.skills = [
            PhysicalNormalAttackBase(
                name = 'Oceanborne',
                cost = PhysicalNormalAttackBase.get_cost(self.element),
            ),
            Tidecaller(),
            Wavestrider(),
            Stormbreaker(),
        ]



from enum import Enum
from ..utils import BaseModel
from typing import Literal, List
from .interaction import (
    ChooseCharactorResponse,
    RerollDiceResponse,
    DeclareRoundEndResponse,
    SwitchCharactorResponse,
)
from .consts import DieColor, ObjectType, SkillType
from .modifiable_values import DamageValue
from .struct import MultipleObjectPosition, ObjectPosition


class ActionTypes(str, Enum):
    EMPTY = 'EMPTY'
    DRAW_CARD = 'DRAW_CARD'
    RESTORE_CARD = 'RESTORE_CARD'
    REMOVE_CARD = 'REMOVE_CARD'
    CHOOSE_CHARACTOR = 'CHOOSE_CHARACTOR'
    CREATE_DICE = 'CREATE_DICE'
    REMOVE_DICE = 'REMOVE_DICE'
    DECLARE_ROUND_END = 'DECLARE_ROUND_END'
    ACTION_END = 'ACTION_END'
    SWITCH_CHARACTOR = 'SWITCH_CHARACTOR'
    CHARGE = 'CHARGE'
    USE_SKILL = 'USE_SKILL'
    USE_CARD = 'USE_CARD'
    SKILL_END = 'SKILL_END'
    CHARACTOR_DEFEATED = 'CHARACTOR_DEFEATED'
    CHARACTOR_REVIVE = 'CHARACTOR_REVIVE'
    CREATE_OBJECT = 'CREATE_OBJECT'
    REMOVE_OBJECT = 'REMOVE_OBJECT'
    OBJECT_REMOVED = 'OBJECT_REMOVED'
    CHANGE_OBJECT_USAGE = 'CHANGE_OBJECT_USAGE'
    MOVE_OBJECT = 'MOVE_OBJECT'
    CONSUME_ARCANE_LEGEND = 'CONSUME_ARCANE_LEGEND'

    # system phase events
    GAME_START = 'GAME_START'
    ROUND_PREPARE = 'ROUND_PREPARE'
    PLAYER_ACTION_START = 'PLAYER_ACTION_START'
    ROUND_END = 'ROUND_END'

    # make damage related events
    RECEIVE_DAMAGE = 'RECEIVE_DAMAGE'
    MAKE_DAMAGE = 'MAKE_DAMAGE'
    AFTER_MAKE_DAMAGE = 'AFTER_MAKE_DAMAGE'

    # generate request actions
    GENERATE_CHOOSE_CHARACTOR = 'GENERATE_CHOOSE_CHARACTOR'
    GENERATE_REROLL_DICE = 'GENERATE_REROLL_DICE'
    GENERATE_SWITCH_CARD = 'GENERATE_SWITCH_CARD'

    # change game state actions
    SKIP_PLAYER_ACTION = 'SKIP_PLAYER_ACTION'


class ActionBase(BaseModel):
    """
    Base class for game actions.
    An action contains arguments to make changes to the game table.

    Attributes:
        action_type (Literal[ActionTypes]): The type of the action.
        record_level (int): The level of the action to record in match.history,
            lower level means more important. It should be a positive number.
    """
    type: Literal[ActionTypes.EMPTY] = ActionTypes.EMPTY

    record_level: int = 100


class DrawCardAction(ActionBase):
    """
    Action for drawing cards.
    """
    type: Literal[ActionTypes.DRAW_CARD] = ActionTypes.DRAW_CARD
    record_level: int = 10
    player_idx: int
    number: int
    blacklist_names: List[str] = []
    whitelist_names: List[str] = []
    blacklist_types: List[ObjectType] = []
    whitelist_types: List[ObjectType] = []
    blacklist_cost_labels: int = 0
    whitelist_cost_labels: int = 0
    draw_if_filtered_not_enough: bool


class RestoreCardAction(ActionBase):
    """
    Action for restoring cards.
    """
    type: Literal[ActionTypes.RESTORE_CARD] = ActionTypes.RESTORE_CARD
    player_idx: int
    card_idxs: List[int]


class RemoveCardAction(ActionBase):
    """
    Action for removing cards.
    """
    type: Literal[ActionTypes.REMOVE_CARD] = ActionTypes.REMOVE_CARD
    record_level: int = 20
    position: ObjectPosition
    remove_type: Literal['USED', 'BURNED']


class ChooseCharactorAction(ActionBase):
    """
    Action for choosing charactors.
    """
    type: Literal[ActionTypes.CHOOSE_CHARACTOR] = ActionTypes.CHOOSE_CHARACTOR
    record_level: int = 10
    player_idx: int
    charactor_idx: int

    @classmethod
    def from_response(cls, response: ChooseCharactorResponse):
        """
        Generate ChooseCharactorAction from ChooseCharactorResponse.
        """
        return cls(
            player_idx = response.player_idx,
            charactor_idx = response.charactor_idx,
        )


# 5


class CreateDiceAction(ActionBase):
    """
    Action for creating dice.

    Args:
        player_idx (int): The index of the player to create the dice for.
        number (int): The number of dice to create.
        color (DieColor | None): The color of the dice to create. If None,
            the following generate rules will be activated.
        random (bool): Whether to randomly generate the color of dice.
        different (bool): Whether to generate different colors of dice.
    """
    type: Literal[ActionTypes.CREATE_DICE] = ActionTypes.CREATE_DICE
    player_idx: int
    number: int
    color: DieColor | None = None
    random: bool = False
    different: bool = False


class RemoveDiceAction(ActionBase):
    """
    Action for removing dice.

    Args:
        player_idx (int): The index of the player to remove the dice for.
        dice_idxs (List[int]): The indices of the dice to remove.
    """
    type: Literal[ActionTypes.REMOVE_DICE] = ActionTypes.REMOVE_DICE
    player_idx: int
    dice_idxs: List[int]

    @classmethod
    def from_response(cls, response: RerollDiceResponse):
        """
        Generate RemoveDiceAction from RerollDiceResponse.
        """
        return cls(
            player_idx = response.player_idx,
            dice_idxs = response.reroll_dice_idxs,
        )


class DeclareRoundEndAction(ActionBase):
    """
    Action for declaring the end of the round.
    """
    type: Literal[ActionTypes.DECLARE_ROUND_END] = \
        ActionTypes.DECLARE_ROUND_END
    record_level: int = 10
    player_idx: int

    @classmethod
    def from_response(cls, response: DeclareRoundEndResponse):
        """
        Generate DeclareRoundEndAction from DeclareRoundEndResponse.
        """
        return cls(
            player_idx = response.player_idx,
        )


class ActionEndAction(ActionBase):
    """
    Action end, and if is a combat action, change current player.
    the position means the action source, i.e. the skill,
    or the charactor who switch out.
    """
    type: Literal[ActionTypes.ACTION_END] = ActionTypes.ACTION_END
    action_label: int  # Refer to PlayerActionLabels
    do_combat_action: bool
    position: ObjectPosition


class SwitchCharactorAction(ActionBase):
    """
    Action for switching charactor.
    """
    type: Literal[ActionTypes.SWITCH_CHARACTOR] = ActionTypes.SWITCH_CHARACTOR
    record_level: int = 10
    player_idx: int
    charactor_idx: int

    @classmethod
    def from_response(cls, response: SwitchCharactorResponse):
        """
        Generate SwitchCharactorAction from SwitchCharactorResponse.
        """
        return cls(
            player_idx = response.player_idx,
            charactor_idx = response.request.target_charactor_idx
        )


# 10


class MakeDamageAction(ActionBase):
    """
    Action for making damage. Heal treats as negative damage. Elemental 
    applies to the charactor (e.g. Kokomi) treats as zero damage.

    Args:
        source_player_idx (int): The index of the player to make damage from.
        damage_value_list (List[DamageValue]): The damage values to make.
        target_player_idx (int): The index of the player to make damage to.
        do_charactor_change (bool): Whether to change charactor after making
            damage.
        charactor_change_idx (int): The charactor index of the charactor who 
            will be changed to. It should not be a defeated charactor.
        TODO: is these two player idx necessary? Can a damage made to both
        players? (e.g. Kokomi's elemental burst)
    """
    type: Literal[ActionTypes.MAKE_DAMAGE] = ActionTypes.MAKE_DAMAGE
    record_level: int = 10
    source_player_idx: int
    damage_value_list: List[DamageValue]
    target_player_idx: int

    # charactor change
    do_charactor_change: bool = False
    charactor_change_idx: int = -1

    def __init__(self, *argv, **kwargs):
        super().__init__(*argv, **kwargs)
        for damage_value in self.damage_value_list:
            assert damage_value.position.id == self.damage_value_list[
                0].position.id, 'all damage should from same source'


class ChargeAction(ActionBase):
    """
    Action for charging.
    """
    type: Literal[ActionTypes.CHARGE] = ActionTypes.CHARGE
    player_idx: int
    charactor_idx: int
    charge: int


class UseSkillAction(ActionBase):
    """
    Action for using skill.
    """
    type: Literal[ActionTypes.USE_SKILL] = ActionTypes.USE_SKILL
    record_level: int = 10
    skill_position: ObjectPosition


class UseCardAction(ActionBase):
    """
    Action for using card.
    """
    type: Literal[ActionTypes.USE_CARD] = ActionTypes.USE_CARD
    record_level: int = 10
    card_position: ObjectPosition
    target: ObjectPosition | MultipleObjectPosition | None


class SkillEndAction(ActionBase):
    """
    Action for ending skill.
    """
    type: Literal[ActionTypes.SKILL_END] = ActionTypes.SKILL_END
    position: ObjectPosition
    target_position: ObjectPosition  # always initial active charactor now
    skill_type: SkillType


# 15


class CharactorDefeatedAction(ActionBase):
    """
    Action for charactor defeated.
    """
    type: Literal[ActionTypes.CHARACTOR_DEFEATED] = \
        ActionTypes.CHARACTOR_DEFEATED
    player_idx: int
    charactor_idx: int


class CreateObjectAction(ActionBase):
    """
    Action for creating objects, e.g. status, summons, supports.
    Note some objects are not created but moved, e.g. equipments and supports,
    for these objects, do not use this action.
    """
    type: Literal[ActionTypes.CREATE_OBJECT] = \
        ActionTypes.CREATE_OBJECT
    object_position: ObjectPosition
    object_name: str
    object_arguments: dict


class RemoveObjectAction(ActionBase):
    """
    Action for removing objects. 

    Args:
        object_position (ObjectPosition): The ID and position of the 
            object to remove.
    """
    type: Literal[ActionTypes.REMOVE_OBJECT] = \
        ActionTypes.REMOVE_OBJECT
    object_position: ObjectPosition


class ChangeObjectUsageAction(ActionBase):
    """
    Action for changing object usage.
    """
    type: Literal[ActionTypes.CHANGE_OBJECT_USAGE] = \
        ActionTypes.CHANGE_OBJECT_USAGE
    object_position: ObjectPosition
    change_type: Literal['DELTA', 'SET']
    change_usage: int
    min_usage: int = 0
    max_usage: int = 999


class MoveObjectAction(ActionBase):
    """
    Action for moving objects.
    """
    type: Literal[ActionTypes.MOVE_OBJECT] = \
        ActionTypes.MOVE_OBJECT
    record_level: int = 10
    object_position: ObjectPosition
    target_position: ObjectPosition
    # for Master of Weaponry etc. in 4.1, when mark true, reset round usage
    reset_usage: bool = False


# 20


class ConsumeArcaneLegendAction(ActionBase):
    """
    Action for consuming arcane legend.
    """
    type: Literal[ActionTypes.CONSUME_ARCANE_LEGEND] = \
        ActionTypes.CONSUME_ARCANE_LEGEND
    player_idx: int


class GenerateChooseCharactorRequestAction(ActionBase):
    """
    Action for generating choose charactor action.
    """
    type: Literal[ActionTypes.GENERATE_CHOOSE_CHARACTOR] = \
        ActionTypes.GENERATE_CHOOSE_CHARACTOR
    player_idx: int


class GenerateRerollDiceRequestAction(ActionBase):
    """
    Action for generating reroll dice request.
    """
    type: Literal[ActionTypes.GENERATE_REROLL_DICE] = \
        ActionTypes.GENERATE_REROLL_DICE
    player_idx: int
    reroll_times: int


class SkipPlayerActionAction(ActionBase):
    """
    Action for skipping current player action.
    """
    type: Literal[ActionTypes.SKIP_PLAYER_ACTION] = \
        ActionTypes.SKIP_PLAYER_ACTION


class CharactorReviveAction(ActionBase):
    """
    Action for charactor revive.
    """
    type: Literal[ActionTypes.CHARACTOR_REVIVE] = ActionTypes.CHARACTOR_REVIVE
    record_level: int = 10
    player_idx: int
    charactor_idx: int
    revive_hp: int


# 25


class GenerateSwitchCardRequestAction(ActionBase):
    """
    Action for generating switch card request.
    """
    type: Literal[ActionTypes.GENERATE_SWITCH_CARD] = \
        ActionTypes.GENERATE_SWITCH_CARD
    player_idx: int


Actions = (
    ActionBase 
    | DrawCardAction 
    | RestoreCardAction 
    | RemoveCardAction
    | ChooseCharactorAction
    # 5
    | CreateDiceAction
    | RemoveDiceAction
    | DeclareRoundEndAction
    | ActionEndAction
    | SwitchCharactorAction
    # 10
    | MakeDamageAction
    | ChargeAction
    | UseSkillAction
    | UseCardAction
    | SkillEndAction
    # 15
    | CharactorDefeatedAction
    | CreateObjectAction
    | RemoveObjectAction
    | ChangeObjectUsageAction
    | MoveObjectAction
    # 20
    | ConsumeArcaneLegendAction
    | GenerateChooseCharactorRequestAction
    | GenerateRerollDiceRequestAction
    | SkipPlayerActionAction
    | CharactorReviveAction

)



from typing import Any, List, Literal

from ...summon.base import AOESummonBase

from ...action import Actions
from ...struct import Cost

from ...consts import (
    DamageElementalType, DieColor, ElementType, FactionType, 
    WeaponType
)
from ..charactor_base import (
    AOESkillBase, ElementalBurstBase, ElementalNormalAttackBase, 
    ElementalSkillBase, PhysicalNormalAttackBase, CharactorBase, SkillTalent
)


# Summons


class SacredCryoPearl(AOESummonBase):
    name: Literal['Sacred Cryo Pearl'] = 'Sacred Cryo Pearl'
    desc: str = (
        'End Phase: Deal 1 Cryo DMG, deal 1 Piercing DMG to all opposing '
        'characters on standby.'
    )
    version: Literal['3.3'] = '3.3'
    usage: int = 2
    max_usage: int = 2
    damage_elemental_type: DamageElementalType = DamageElementalType.CRYO
    damage: int = 1
    back_damage: int = 1


# Skills


class TrailOftheQilin(ElementalSkillBase):
    name: Literal['Trail of the Qilin'] = 'Trail of the Qilin'
    desc: str = '''Deals 1 Cryo DMG, creates 1 Ice Lotus.'''
    damage: int = 1
    damage_type: DamageElementalType = DamageElementalType.CRYO
    cost: Cost = Cost(
        elemental_dice_color = DieColor.CRYO,
        elemental_dice_number = 3
    )

    def get_actions(self, match: Any) -> List[Actions]:
        """
        Attack and create object
        """
        return super().get_actions(match) + [
            self.create_team_status('Ice Lotus'),
        ]


class FrostflakeArrow(ElementalNormalAttackBase, AOESkillBase):
    name: Literal['Frostflake Arrow'] = 'Frostflake Arrow'
    desc: str = '''Deals 1 Cryo DMG, creates 1 Ice Lotus.'''
    damage: int = 2
    damage_type: DamageElementalType = DamageElementalType.CRYO
    back_damage: int = 2
    cost: Cost = Cost(
        elemental_dice_color = DieColor.CRYO,
        elemental_dice_number = 5
    )

    use_counter: int = 0

    def get_actions(self, match: Any) -> List[Actions]:
        """
        Attack, if has talent, back damage increase
        """
        self.use_counter += 1
        charactor = match.player_tables[self.position.player_idx].charactors[
            self.position.charactor_idx]
        if charactor.talent is not None:
            # if have talent, may increase damage
            if self.use_counter >= 2:
                # condition satisfied
                self.back_damage = 3
                if charactor.talent.version == '3.3':
                    # if is old talent
                    self.damage = 3
            else:
                self.back_damage = 2
                self.damage = 2
        return super().get_actions(match)


class CelestialShower(ElementalBurstBase, AOESkillBase):
    name: Literal['Celestial Shower'] = 'Celestial Shower'
    desc: str = (
        'Deals 2 Cryo DMG, deals 1 Piercing DMG to all opposing characters on '
        'standby, summons 1 Sacred Cryo Pearl.'
    )
    damage: int = 2
    damage_type: DamageElementalType = DamageElementalType.CRYO
    back_damage: int = 1
    cost: Cost = Cost(
        elemental_dice_color = DieColor.CRYO,
        elemental_dice_number = 3,
        charge = 3
    )

    def get_actions(self, match: Any) -> List[Actions]:
        return super().get_actions(match) + [
            self.create_summon('Sacred Cryo Pearl')
        ]


# Talents


class UndividedHeart(SkillTalent):
    name: Literal['Undivided Heart']
    desc: str = (
        'Combat Action: When your active character is Ganyu, equip this card. '
        'After Ganyu equips this card, immediately use Frostflake Arrow once. '
        'When your Ganyu, who has this card equipped, uses Frostflake Arrow: '
        'If this Skill has been used before during this match, the Piercing '
        'DMG dealt to all opposing characters on standby is changed to 3.'
    )
    version: Literal['3.7'] = '3.7'
    charactor_name: Literal['Ganyu'] = 'Ganyu'
    cost: Cost = Cost(
        elemental_dice_color = DieColor.CRYO,
        elemental_dice_number = 5,
    )
    skill: FrostflakeArrow = FrostflakeArrow()


# charactor base


class Ganyu(CharactorBase):
    name: Literal['Ganyu']
    version: Literal['3.7'] = '3.7'
    desc: str = '''"Plenilune Gaze" Ganyu'''
    element: ElementType = ElementType.CRYO
    max_hp: int = 10
    max_charge: int = 3
    skills: List[
        PhysicalNormalAttackBase | TrailOftheQilin | FrostflakeArrow 
        | CelestialShower
    ] = []
    faction: List[FactionType] = [
        FactionType.LIYUE
    ]
    weapon_type: WeaponType = WeaponType.BOW
    talent: UndividedHeart | None = None

    def _init_skills(self) -> None:
        self.skills = [
            PhysicalNormalAttackBase(
                name = 'Liutian Archery',
                cost = PhysicalNormalAttackBase.get_cost(self.element),
            ),
            TrailOftheQilin(),
            FrostflakeArrow(),
            CelestialShower()
        ]
